import collections
import collections.abc
import sys
import typing
import warnings
from asyncio import iscoroutinefunction
from enum import Enum
from inspect import (
    isclass,
    ismethod,
)
from json import loads as json_loads
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Concatenate,
    Generic,
    Optional,
    ParamSpec,
    Protocol,
    Type,
    TypedDict,
    TypeVar,
    get_args,
    get_origin,
)

import starlette.responses
from fastapi.responses import JSONResponse, Response
from inflection import camelize
from pydantic import BaseModel, Field, create_model
from pydantic.fields import FieldInfo

from mountaineer.annotation_helpers import MountaineerUnsetValue
from mountaineer.exceptions import APIException, RequestValidationError
from mountaineer.render import FieldClassDefinition, Metadata, RenderBase, RenderNull

if sys.version_info >= (3, 11):
    from typing import NotRequired
else:
    from typing_extensions import NotRequired

if TYPE_CHECKING:
    from mountaineer.controller import ControllerBase


class FunctionActionType(Enum):
    RENDER = "render"
    SIDEEFFECT = "sideeffect"
    PASSTHROUGH = "passthrough"


class ResponseModelType(Enum):
    SINGLE_RESPONSE = "SINGLE_RESPONSE"
    ITERATOR_RESPONSE = "ITERATOR_RESPONSE"


P = TypeVar("P")

if sys.version_info >= (3, 11):

    class SideeffectResponseBase(TypedDict, Generic[P]):
        passthrough: P
        # We can't yet typehint across the boundary of the sideeffect -> render
        # within one class since @sideeffect is isolated to just the wrapped function
        sideeffect: NotRequired[Any]
else:
    # Inheriting from both TypedDict and Generic[P] is not supported in Python 3.10 and below
    class SideeffectResponseBase(dict, Generic[P]):
        pass


T = ParamSpec("T")
R = TypeVar("R")
C = TypeVar("C", covariant=True)


class SideeffectWrappedCallable(Protocol[C, T, R]):
    def __call__(
        self: Any, *args: T.args, **kwargs: T.kwargs
    ) -> Awaitable[SideeffectResponseBase[R]]: ...

    # Since the original function is extracted directly from the class, it's
    # just a hanging function and no longer an instance method. Users therefore
    # need to call it with the class instance explicitly as the first argument.
    original: Callable[Concatenate[C, T], Awaitable[R]]


class SideeffectRawCallable(Protocol[C, T, R]):
    def __call__(self: Any, *args: T.args, **kwargs: T.kwargs) -> Awaitable[R]: ...

    original: Callable[Concatenate[C, T], Awaitable[R]]


class FunctionMetadata(BaseModel):
    function_name: str
    action_type: FunctionActionType

    # Specified for sideeffects, where all data shouldn't be update. Limits the
    # update to fields defined in this tuple.
    reload_states: tuple[FieldClassDefinition, ...] | None | MountaineerUnsetValue = (
        MountaineerUnsetValue()
    )

    # Defines the data schema returned from the function that will be included in the
    # response payload sent to the client. This might be used for either passthrough
    # or sideeffect
    passthrough_model: Type[BaseModel] | None | MountaineerUnsetValue = (
        MountaineerUnsetValue()
    )
    exception_models: list[Type[APIException]] = [RequestValidationError]
    media_type: str | None | MountaineerUnsetValue = MountaineerUnsetValue()
    is_raw_response: bool = False

    # Render type, defines the data model that is returned by the render typehint
    # If "None", the user has explicitly stated that no render model is returned
    render_model: Type[RenderBase] | None | MountaineerUnsetValue = (
        MountaineerUnsetValue()
    )

    # Inserted when concrete controllers are mounted to the application controller
    return_models: dict[Any, Type[BaseModel]] = Field(default_factory=dict)

    # An API action might be mounted by multiple controllers, if it comes from a superclass
    # that's inherited by multiple child controllers. This lookup lets us track which
    # URL is associated with which controller.
    controller_mounts: dict[Any, str] = Field(default_factory=dict)

    model_config = {
        "arbitrary_types_allowed": True,
    }

    #
    # Accessors for polymorphic variables
    # These convenience methods are used to ensure client callers that this function metadata
    # does have the expected values. They should only be used in cases that you know based on runtime
    # guarantees that certain functions will have certain attributes.
    #
    def get_reload_states(self) -> tuple[FieldClassDefinition, ...] | None:
        if isinstance(self.reload_states, MountaineerUnsetValue):
            raise ValueError("Reload states not set")
        return self.reload_states

    def get_render_model(self) -> Type[RenderBase] | None:
        if isinstance(self.render_model, MountaineerUnsetValue):
            raise ValueError("Render model not set")
        return self.render_model or RenderNull

    def get_passthrough_model(self) -> Type[BaseModel] | None:
        if isinstance(self.passthrough_model, MountaineerUnsetValue):
            raise ValueError("Passthrough model not set")
        return self.passthrough_model

    def get_exception_models(self) -> list[Type[APIException]] | None:
        if isinstance(self.exception_models, MountaineerUnsetValue):
            raise ValueError("Exception models not set")
        return self.exception_models

    def get_media_type(self) -> str | None:
        if isinstance(self.media_type, MountaineerUnsetValue):
            raise ValueError("Media type not set")
        return self.media_type

    def get_is_raw_response(self) -> bool:
        return self.is_raw_response

    def get_return_model(self, controller: Type["ControllerBase"]) -> Type[BaseModel]:
        if controller not in self.return_models:
            raise ValueError("Return model not set")
        return self.return_models[controller]

    def register_controller_url(self, controller: Type["ControllerBase"], url: str):
        if controller in self.controller_mounts:
            # See if it's the same URL
            if self.controller_mounts[controller] != url:
                raise ValueError(
                    f"Controller {controller} already mounted at {self.controller_mounts[controller]} with different URL\n"
                    f"Old: {self.controller_mounts[controller]} New: {url}"
                )
        self.controller_mounts[controller] = url

    def register_return_model(
        self, controller: Type["ControllerBase"], model: Type[BaseModel]
    ):
        if controller in self.return_models:
            if self.return_models[controller] != model:
                raise ValueError(
                    f"Controller {controller} already registered with a different return model\n"
                    f"Old: {self.return_models[controller]} New: {model}"
                )
        self.return_models[controller] = model

    # def get_openapi_extras(self):
    #     openapi_extra: dict[str, Any] = {"is_raw_response": self.get_is_raw_response()}

    #     if not self.get_is_raw_response():
    #         # Pass along relevant tags in the OpenAPI meta struct
    #         # This will appear in the root key of the API route, at the same level of "summary" and "parameters"
    #         if self.get_media_type():
    #             openapi_extra["media_type"] = self.get_media_type()

    #     return openapi_extra


METADATA_ATTRIBUTE = "_mountaineer_metadata"


def init_function_metadata(fn: Callable, action_type: FunctionActionType):
    if ismethod(fn):
        fn = fn.__func__  # type: ignore

    function_name = fn.__name__
    metadata = FunctionMetadata(function_name=function_name, action_type=action_type)
    setattr(fn, METADATA_ATTRIBUTE, metadata)
    return metadata


def get_function_metadata(fn: Callable) -> FunctionMetadata:
    if ismethod(fn):
        fn = fn.__func__  # type: ignore

    if not hasattr(fn, METADATA_ATTRIBUTE):
        raise AttributeError(f"Function {fn.__name__} does not have metadata")

    metadata = getattr(fn, METADATA_ATTRIBUTE)

    if not isinstance(metadata, FunctionMetadata):
        raise AttributeError(f"Function {fn.__name__} has invalid metadata")

    return metadata


def annotation_is_metadata(annotation: type | None):
    if not annotation:
        return

    return annotation == Metadata or annotation == Optional[Metadata]


def fuse_metadata_to_response_typehint(
    metadata: FunctionMetadata,
    controller: "ControllerBase",
    render_model: Type[RenderBase] | None,
) -> Type[BaseModel]:
    """
    Functions can either be marked up with side effects, explicit responses, or both.
    This function merges them into the expected output payload so we can typehint the responses.

    """
    # Prefer to use existing BaseModels where possible, we only create synthetic values
    # if we need to mask the response model
    passthrough_model: Type[BaseModel] | None = None
    sideeffect_model: Type[BaseModel] | None = None

    base_response_name = camelize(metadata.function_name) + "ResponseWrapped"
    base_response_params = {}

    # Prefer the location of the concrete controller when defining the model, since a render function
    # could be imported by multiple controllers
    base_module = controller.__module__

    if metadata.passthrough_model is not None and not isinstance(
        metadata.passthrough_model, MountaineerUnsetValue
    ):
        passthrough_model = metadata.passthrough_model

    if metadata.action_type == FunctionActionType.SIDEEFFECT and render_model:
        # By default, reload all fields
        sideeffect_model = render_model

        if metadata.reload_states is not None and not isinstance(
            metadata.reload_states, MountaineerUnsetValue
        ):
            # Make sure this class actually aligns to the response model
            # If not the user mis-specified the reload states
            #
            # We allow the reload state to be a subclass of the render model, in case the method
            # was originally defined in the superclass. This will result in us sending a subset
            # of the child controller's fields - but since our differential update is based on the
            # original state and modified, this will resolve correctly for the client
            reload_classes = {field.root_model for field in metadata.reload_states}
            reload_keys = {field.key for field in metadata.reload_states}
            if len(reload_classes) != 1 or not issubclass(
                render_model, next(iter(reload_classes))
            ):
                raise ValueError(
                    f"Reload states {reload_classes} do not align to response model {render_model}"
                )
            sideeffect_model = create_model(
                base_response_name + "SideEffectWrapped",
                __module__=base_module,
                **{
                    field_name: (field_definition.annotation, field_definition)  # type: ignore
                    for field_name, field_definition in render_model.model_fields.items()
                    if field_name in reload_keys
                },
            )

    if passthrough_model:
        base_response_params["passthrough"] = (
            passthrough_model,
            FieldInfo(alias="passthrough"),
        )

    if sideeffect_model:
        base_response_params["sideeffect"] = (
            sideeffect_model,
            FieldInfo(alias="sideeffect"),
        )

    model: Type[BaseModel] = create_model(
        base_response_name,
        __module__=base_module,
        **base_response_params,  # type: ignore
    )

    return model


def format_final_action_response(dict_payload: dict[str, Any]):
    """
    Wrapper to allow actions to respond with an explicit JSONResponse, or a dictionary. This lets
    both sideeffects and passthrough payloads to inject header metadata that otherwise can't be captured
    in a regular BaseModel.

    Since the eventual result of an action is a combined sideeffect+passthrough payload, we need to
    merge the final payload into the explicit response.

    """
    responses: list[tuple[str, JSONResponse]] = [
        (key, response)
        for key, response in dict_payload.items()
        if isinstance(response, JSONResponse)
    ]

    if len(responses) > 1:
        raise ValueError(f"Multiple conflicting responses returned: {responses}")

    if len(responses) == 0:
        return dict_payload

    response_key, response = responses[0]
    assert isinstance(response, Response)
    assert isinstance(response.body, bytes)

    dict_payload[response_key] = json_loads(response.body)

    # Now inject the newly formatted response into the response object
    return JSONResponse(
        content=dict_payload,
        status_code=response.status_code,
        headers={
            key: value
            for key, value in response.headers.items()
            if key not in {"content-length", "content-type"}
        },
    )


def extract_response_model_from_signature(
    func: Callable, explicit_response: Type[BaseModel] | None = None
):
    typehinted_response = func.__annotations__.get("return", MountaineerUnsetValue())
    if explicit_response:
        warnings.warn(
            (
                "The response_model argument is deprecated. Instead, just typehint your function explicitly:\n"
                "def my_function() -> MyResponseModel:"
            ),
            DeprecationWarning,
            stacklevel=2,
        )

    if explicit_response:
        return explicit_response, ResponseModelType.SINGLE_RESPONSE

    if isinstance(typehinted_response, MountaineerUnsetValue):
        # This will be converted into a ValueError in the future
        # For now, backwards compatible with the old markup
        warnings.warn(
            (
                f"You must typehint the return value of your `{func}` with either a BaseModel or None.\n"
                "We will stop inferring `None` as a response model in the future."
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return None, ResponseModelType.SINGLE_RESPONSE

    return extract_model_from_decorated_types(typehinted_response)


def extract_model_from_decorated_types(
    type_hint: Any,
) -> tuple[Type[BaseModel] | None, ResponseModelType]:
    """
    Support response_model typehints like Iterator[Type[BaseModel]] and AsyncIterator[Type[BaseModel]].

    """
    origin_type = get_origin(type_hint)

    if type_hint is None:
        return None, ResponseModelType.SINGLE_RESPONSE
    elif isclass(type_hint) and issubclass(type_hint, BaseModel):
        return type_hint, ResponseModelType.SINGLE_RESPONSE
    elif origin_type in (
        typing.Iterator,
        typing.AsyncIterator,
        # At runtime our types are sometimes instantiated as collections.abc objects
        collections.abc.Iterator,
        collections.abc.AsyncIterator,
    ):
        args = get_args(type_hint)
        if args and issubclass(args[0], BaseModel):
            return args[0], ResponseModelType.ITERATOR_RESPONSE
        raise ValueError(
            f"Invalid response_model typehint for iterator action: {type_hint} {origin_type} {args}"
        )
    elif isclass(type_hint) and issubclass(type_hint, starlette.responses.Response):
        # No pydantic model to include in the API schema, instead the endpoint
        # will just return the raw value
        return None, ResponseModelType.SINGLE_RESPONSE

    raise ValueError(
        f"Invalid response_model typehint for standard action: {type_hint}"
    )


def create_original_fn(fn):
    """
    To fulfill our typehints for "original" that are returned from @sideeffect
    and @passthrough we have to make them all async even if they weren't originally.

    """
    if iscoroutinefunction(fn):
        return fn
    else:

        async def async_fn(*args, **kwargs):
            return fn(*args, **kwargs)

        return async_fn



from tqdm import tqdm


def calculate_primes(up_to: int):
    def is_prime(n: int) -> bool:
        if n < 2:
            return False
        for i in range(2, int(n**0.5) + 1):
            if n % i == 0:
                return False
        return True

    return sum(is_prime(i) for i in tqdm(range(2, up_to + 1)))



from dataclasses import dataclass
from typing import Any

from mountaineer.client_builder.interface_builders.base import InterfaceBase
from mountaineer.client_builder.parser import (
    EnumWrapper,
)
from mountaineer.client_builder.typescript import (
    TSLiteral,
    python_payload_to_typescript,
)


@dataclass
class EnumInterface(InterfaceBase):
    name: str
    body: str
    include_export: bool = True

    @classmethod
    def from_enum(cls, enum: EnumWrapper):
        fields: dict[str, Any] = {}

        # Mirror the format of JS enums
        # https://www.typescriptlang.org/docs/handbook/enums.html
        # enum Direction {
        #  Up = 1,
        #  Down,
        #  Left,
        #  Right,
        # }
        for name, value in enum.enum.__members__.items():
            fields[TSLiteral(name)] = value.value

        return cls(
            name=enum.name.global_name,
            # Optional spacing, but make for better enum definitions (A = 'A')
            body=python_payload_to_typescript(fields, dict_equality=" ="),
        )

    def to_js(self) -> str:
        schema_def = f"enum {self.name}"

        # Body includes the { } tokens
        schema_def += f" {self.body}"

        if self.include_export:
            schema_def = f"export {schema_def}"

        return schema_def



import asyncio
from typing import AsyncIterator

from mountaineer import ControllerBase, passthrough
from pydantic import BaseModel


class StreamActionResponse(BaseModel):
    value: str


class StreamController(ControllerBase):
    url = "/stream"
    view_path = "/app/stream/page.tsx"

    async def render(self) -> None:
        pass

    @passthrough
    async def stream_action(self) -> AsyncIterator[StreamActionResponse]:
        for i in range(10):
            yield StreamActionResponse(value=f"streaming {i}\n")
            await asyncio.sleep(1)



import asyncio
import socket
from threading import Thread
from time import time
from typing import Optional

from fastapi import FastAPI
from uvicorn import Config
from uvicorn.server import Server


class UvicornThread(Thread):
    def __init__(
        self,
        *,
        name: str,
        emoticon: str,
        app: FastAPI,
        host: str,
        port: int,
        log_level: str = "info",
        use_logs: bool = True,
    ):
        super().__init__(daemon=True)
        self.app = app
        self.port = port
        self.host = host
        self.log_level = log_level
        self.name = name
        self.emoticon = emoticon
        self.server: Optional[Server] = None
        self.use_logs = use_logs

    def run(self) -> None:
        loop = asyncio.new_event_loop()
        config = Config(
            app=self.app,
            host=self.host,
            port=self.port,
            reload=False,
            access_log=False,
            loop="asyncio",
            log_level=self.log_level,
        )

        server = Server(config)
        self.server = server

        loop.run_until_complete(server.serve())

    async def astart(self, timeout: int = 5) -> None:
        super().start()

        # Wait until the server has self-flagged server.started and it's bound to the
        # desired port. If we we timeout waiting for either signal, then raise an error.
        did_start = False
        start = time()
        while time() - start < timeout:
            is_mounted = (
                self.server and self.server.started and not self._is_port_free()
            )
            if is_mounted:
                did_start = True
                break

            await asyncio.sleep(0.1)

        if not did_start:
            raise TimeoutError(
                f"Server did not start in {timeout}s (checked {self.host}:{self.port})"
            )

    async def astop(self, timeout: int = 1) -> None:
        """
        Attempts to stop the server gracefully. If the server does not stop
        within the given timeout, then we forcefully kill the server.

        """
        if self.server is not None:
            self.server.should_exit = True

        # Check if the port is still bound
        did_stop = False
        start = time()
        while time() - start < timeout:
            if self._is_port_free():
                did_stop = True
                break

            await asyncio.sleep(0.1)

        # If we get here, the port is still bound after all our checks
        if not did_stop:
            raise TimeoutError(f"Server did not stop after {timeout}s")

    def start(self) -> None:
        raise NotImplementedError("Use astart() instead")

    def stop(self) -> None:
        raise NotImplementedError("Use astop() instead")

    def _is_port_free(self) -> bool:
        """
        Check if the port is free by attempting to bind to it.
        Returns True if the port is free, False if it's still in use.
        """
        try:
            # Create a socket and try to bind to the port
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.settimeout(0.1)  # Set a short timeout
                s.bind((self.host, self.port))
                # If we get here, binding succeeded, port is free
                return True
        except (socket.error, OSError):
            # Port is still in use
            return False



import logging
from contextlib import contextmanager
from json import dumps as json_dumps
from logging import Formatter, StreamHandler, getLogger
from os import environ
from pathlib import Path
from tempfile import mkdtemp
from time import monotonic_ns, time

from click import secho

VERBOSITY_MAPPING = {
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
}


class JsonFormatter(Formatter):
    def format(self, record):
        log_record = {
            "level": record.levelname,
            "name": record.name,
            "timestamp": self.formatTime(record, self.datefmt),
            "message": record.getMessage(),
        }
        if record.exc_info:
            log_record["exception"] = self.formatException(record.exc_info)
        return json_dumps(log_record)


class ColorHandler(StreamHandler):
    def emit(self, record):
        try:
            msg = self.format(record)
            if record.levelno == logging.WARNING:
                secho(msg, fg="yellow")
            elif record.levelno >= logging.ERROR:
                secho(msg, fg="red")
            else:
                secho(msg)
        except Exception:
            self.handleError(record)


def setup_logger(name, log_level=logging.DEBUG):
    """
    Constructor for the main logger used by Mountaineer and optionally for client
    applications as well. Provided convenient defaults for log level and formatting,
    alongside coloring of stdout/stderr messages and JSON fields for structured parsing.

    Logs are formatted one per line:
    ```json
    {"level": "INFO", "name": "myapp.logging", "timestamp": "2025-02-25 20:40:35,896", "message": "Application started"}
    ```

    To grep over these logs and filter for level, do:

    ```bash
    # Filter for specific log level
    grep '"level": "ERROR"' logfile.txt

    # Filter logs by service name
    grep '"name": "myapp.logging"' logfile.txt

    # With jq for more advanced JSON parsing
    cat logfile.txt | jq 'select(.level=="ERROR" or .level=="WARNING")'

    # Filter by message content
    grep -E '"message": ".*database.*"' logfile.txt
    ```

    ```python {{sticky: True}}
    from mountaineer.logging import setup_logger
    import logging

    # Create a logger for your module
    logger = setup_logger(__name__, log_level=logging.INFO)

    # Use the logger
    logger.info("Application started")
    logger.warning("Configuration incomplete")
    logger.error("Failed to connect to database")
    ```

    :param name: The name of the logger, typically the module name
    :param log_level: The logging level. Defaults to logging.DEBUG to log everything.

    :return: A configured logger instance

    """
    logger = getLogger(name)
    logger.setLevel(log_level)

    # Create a handler that writes log records to the standard error
    handler = ColorHandler()
    handler.setLevel(log_level)

    formatter = JsonFormatter()
    handler.setFormatter(formatter)

    logger.addHandler(handler)

    return logger


@contextmanager
def log_time_duration(message: str):
    """
    Context manager to time a code block at runtime.

    ```python
    with log_time_duration("Long computation"):
        # Simulate work
        sleep(10)
    ```

    """
    start = monotonic_ns()
    yield
    LOGGER.debug(f"{message} : Took {(monotonic_ns() - start) / 1e9:.2f}s")


def setup_internal_logger(name: str):
    """
    Our global logger should only surface warnings and above by default.

    To adjust Mountaineer logging, set the MOUNTAINEER_LOG_LEVEL environment
    variable in your local session. By default it is set to WARNING and above.

    """
    return setup_logger(
        name,
        log_level=VERBOSITY_MAPPING[environ.get("MOUNTAINEER_LOG_LEVEL", "WARNING")],
    )


def pluralize(count: int, singular: str, plural: str) -> str:
    return singular if count == 1 else plural


RUNTIME_ARITFACT_TMP_DIR: Path | None = None


def debug_log_artifact(artifact_prefix: str, extension: str, content: str):
    global RUNTIME_ARITFACT_TMP_DIR

    # Only log during highest level of verbosity
    if (
        VERBOSITY_MAPPING[environ.get("MOUNTAINEER_LOG_LEVEL", "WARNING")]
        > logging.DEBUG
    ):
        return

    if RUNTIME_ARITFACT_TMP_DIR is None:
        RUNTIME_ARITFACT_TMP_DIR = Path(mkdtemp())
        LOGGER.warning(
            f"Created temporary directory for runtime artifacts: {RUNTIME_ARITFACT_TMP_DIR}"
        )

    path = RUNTIME_ARITFACT_TMP_DIR / f"{artifact_prefix}-{time()}.{extension}"
    path.write_text(content)

    return path


def reset_artifact_dir():
    global RUNTIME_ARITFACT_TMP_DIR
    RUNTIME_ARITFACT_TMP_DIR = None


LOGGER = setup_internal_logger(__name__)



from pathlib import Path
from shutil import rmtree as shutil_rmtree

from mountaineer.app import AppController
from mountaineer.client_builder.aliases import AliasManager
from mountaineer.client_builder.file_generators.base import ParsedController
from mountaineer.client_builder.file_generators.globals import (
    GlobalControllerGenerator,
    GlobalLinkGenerator,
)
from mountaineer.client_builder.file_generators.locals import (
    LocalActionGenerator,
    LocalIndexGenerator,
    LocalLinkGenerator,
    LocalModelGenerator,
    LocalUseServerGenerator,
)
from mountaineer.client_builder.parser import (
    ControllerParser,
)
from mountaineer.controller_layout import LayoutControllerBase as LayoutControllerBase
from mountaineer.paths import ManagedViewPath
from mountaineer.static import get_static_path


class APIBuilder:
    """
    Main entrypoint for building the auto-generated typescript code. This includes
    the server provided API used by useServer.
    """

    def __init__(
        self,
        app: AppController,
        live_reload_port: int | None = None,
        build_cache: Path | None = None,
    ):
        self.app = app
        self.live_reload_port = live_reload_port
        self.build_cache = build_cache
        self.view_root = ManagedViewPath.from_view_root(app._view_root)

        self.alias_manager = AliasManager()

        self.update_controller(app)

    def update_controller(self, controller: AppController):
        self.app = controller
        self.view_root = ManagedViewPath.from_view_root(controller._view_root)

    async def build_all(self):
        # Totally clear away the old build cache, so we start fresh
        # and don't have additional files hanging around
        for clear_dir in [
            self.view_root.get_managed_ssr_dir(),
            self.view_root.get_managed_static_dir(),
        ]:
            if clear_dir.exists():
                shutil_rmtree(clear_dir)

        await self.build_use_server()
        # await self.build_fe_diff(None)

    async def build_use_server(self):
        # Parse all controllers first
        parser, parsed_controller = self._parse_all_controllers()
        self._assign_unique_names(parser)

        # Generate all the required files
        self._generate_static_files()
        self._generate_global_files(parsed_controller)
        self._generate_local_files(parsed_controller)

    def _parse_all_controllers(self):
        """Parse all controllers and store their parsed representations"""
        parser = ControllerParser()
        parsed_controllers: list[ParsedController] = []

        for controller_def in self.app.controllers:
            controller = controller_def.controller

            # Parse the controller
            parsed_wrapper = parser.parse_controller(controller.__class__)

            # Get view path
            view_path = self.view_root.get_controller_view_path(controller)

            # Create ParsedController instance
            parsed_controllers.append(
                ParsedController(
                    wrapper=parsed_wrapper,
                    view_path=view_path,
                    url_prefix=controller_def.url_prefix,
                    is_layout=isinstance(controller, LayoutControllerBase),
                )
            )

        return parser, parsed_controllers

    def _assign_unique_names(self, parser: ControllerParser):
        self.alias_manager.assign_global_names(parser)
        self.alias_manager.assign_local_names(parser)

    def _generate_global_files(self, parsed_controllers: list[ParsedController]):
        global_root = self.view_root.get_managed_code_dir()

        global_controller_generator = GlobalControllerGenerator(
            controller_wrappers=[
                controller.wrapper for controller in parsed_controllers
            ],
            managed_path=global_root / "controllers.ts",
        )
        global_link_generator = GlobalLinkGenerator(
            parsed_controllers=parsed_controllers,
            managed_path=global_root / "links.ts",
        )

        global_controller_generator.build()
        global_link_generator.build()

    def _generate_local_files(self, parsed_controllers: list[ParsedController]):
        global_root = self.view_root.get_managed_code_dir()

        for parsed_controller in parsed_controllers:
            managed_path = parsed_controller.view_path.get_managed_code_dir()

            local_link_generator = LocalLinkGenerator(
                controller=parsed_controller.wrapper,
                managed_path=managed_path / "links.ts",
                global_root=global_root,
            )
            local_action_generator = LocalActionGenerator(
                controller=parsed_controller.wrapper,
                managed_path=managed_path / "actions.ts",
                global_root=global_root,
            )
            local_model_generator = LocalModelGenerator(
                controller=parsed_controller.wrapper,
                managed_path=managed_path / "models.ts",
                global_root=global_root,
            )
            local_use_server_generator = LocalUseServerGenerator(
                controller=parsed_controller.wrapper,
                managed_path=managed_path / "useServer.ts",
                global_root=global_root,
            )
            local_index_generator = LocalIndexGenerator(
                controller=parsed_controller.wrapper,
                managed_path=managed_path / "index.ts",
                global_root=global_root,
            )

            # Controller-only Files
            if not parsed_controller.is_layout:
                local_link_generator.build()

            # Controllers + Layout files
            local_action_generator.build()
            local_model_generator.build()
            local_use_server_generator.build()

            # Since the local index generator reads created files on disk to
            # determine what to re-export, it needs to always go after the
            # other generators
            local_index_generator.build()

    def _generate_static_files(self):
        """Copy over static files required for the client"""
        managed_code_dir = self.view_root.get_managed_code_dir()
        for static_file in ["api.ts", "live_reload.ts"]:
            content = get_static_path(static_file).read_text()
            (managed_code_dir / static_file).write_text(content)



from datetime import datetime
from enum import Enum
from typing import Any, Type

import pytest
from pydantic import BaseModel

from mountaineer.__tests__.client_builder.interface_builders.common import (
    create_field_wrapper,
    create_model_wrapper,
)
from mountaineer.__tests__.client_builder.interface_builders.test_enum import (
    create_enum_wrapper,
)
from mountaineer.client_builder.interface_builders.model import ModelInterface
from mountaineer.client_builder.types import Or


class TestBasicInterfaceGeneration:
    def test_simple_model_interface(self):
        class FieldType(Enum):
            STRING = "string"
            NUMBER = "number"

        wrapper = create_model_wrapper(
            BaseModel,
            "SimpleModel",
            [
                create_field_wrapper("string_field", str),
                create_field_wrapper("int_field", int),
                create_field_wrapper("optional_field", Or(str, None), required=False),
                create_field_wrapper("enum_field", create_enum_wrapper(FieldType)),
            ],
        )

        interface = ModelInterface.from_model(wrapper)
        ts_code = interface.to_js()

        assert "export interface SimpleModel" in ts_code
        assert "string_field: string" in ts_code
        assert "int_field: number" in ts_code
        assert "optional_field?: string" in ts_code
        assert "enum_field: FieldType" in ts_code

    def test_no_export(self):
        wrapper = create_model_wrapper(
            BaseModel, "SimpleModel", [create_field_wrapper("field", str)]
        )

        interface = ModelInterface.from_model(wrapper)
        interface.include_export = False
        ts_code = interface.to_js()

        assert not ts_code.startswith("export")
        assert ts_code.startswith("interface SimpleModel")

    @pytest.mark.parametrize(
        "field_name,field_type,expected_ts",
        [
            ("string_field", str, "string"),
            ("int_field", int, "number"),
            ("bool_field", bool, "boolean"),
            ("float_field", float, "number"),
            ("date_field", datetime, "string"),
        ],
    )
    def test_field_type_conversion(
        self, field_name: str, field_type: Type[Any], expected_ts: str
    ):
        wrapper = create_model_wrapper(
            BaseModel, "DynamicModel", [create_field_wrapper(field_name, field_type)]
        )

        interface = ModelInterface.from_model(wrapper)
        assert f"{field_name}: {expected_ts}" in interface.to_js()


class TestInheritanceHandling:
    def test_single_inheritance(self):
        parent_wrapper = create_model_wrapper(
            BaseModel,
            "ParentModel",
            [
                create_field_wrapper("parent_field", str),
                create_field_wrapper("shared_field", int),
            ],
        )

        child_wrapper = create_model_wrapper(
            BaseModel,
            "ChildModel",
            [
                create_field_wrapper("child_field", bool),
                create_field_wrapper("shared_field", float),  # Override type
            ],
            superclasses=[parent_wrapper],
        )

        interface = ModelInterface.from_model(child_wrapper)
        ts_code = interface.to_js()

        assert "interface ChildModel extends ParentModel" in ts_code
        assert "child_field: boolean" in ts_code
        assert "shared_field: number" in ts_code

    def test_multiple_inheritance(self):
        base1_wrapper = create_model_wrapper(
            BaseModel, "MultiInheritBase1", [create_field_wrapper("base1_field", str)]
        )

        base2_wrapper = create_model_wrapper(
            BaseModel, "MultiInheritBase2", [create_field_wrapper("base2_field", int)]
        )

        child_wrapper = create_model_wrapper(
            BaseModel,
            "MultiInheritChild",
            [create_field_wrapper("child_field", bool)],
            superclasses=[base1_wrapper, base2_wrapper],
        )

        interface = ModelInterface.from_model(child_wrapper)
        ts_code = interface.to_js()

        assert "extends MultiInheritBase1, MultiInheritBase2" in ts_code
        assert "child_field: boolean" in ts_code


class TestEdgeCases:
    def test_empty_model(self):
        wrapper = create_model_wrapper(BaseModel, "EmptyModel", [])
        interface = ModelInterface.from_model(wrapper)
        ts_code = interface.to_js()

        assert "interface EmptyModel {\n\n}" in ts_code

    def test_all_optional_fields(self):
        wrapper = create_model_wrapper(
            BaseModel,
            "OptionalModel",
            [
                create_field_wrapper("field1", str, required=False),
                create_field_wrapper("field2", int, required=False),
            ],
        )

        interface = ModelInterface.from_model(wrapper)
        ts_code = interface.to_js()

        assert "field1?: string" in ts_code
        assert "field2?: number" in ts_code



from graphlib import TopologicalSorter
from typing import Any

from inflection import camelize

from mountaineer.client_builder.file_generators.base import (
    CodeBlock,
    FileGeneratorBase,
    ParsedController,
)
from mountaineer.client_builder.interface_builders.controller import ControllerInterface
from mountaineer.client_builder.interface_builders.enum import EnumInterface
from mountaineer.client_builder.interface_builders.exception import ExceptionInterface
from mountaineer.client_builder.interface_builders.model import ModelInterface
from mountaineer.client_builder.parser import (
    ControllerWrapper,
    EnumWrapper,
    ModelWrapper,
)
from mountaineer.client_builder.typescript import (
    TSLiteral,
    python_payload_to_typescript,
)
from mountaineer.paths import ManagedViewPath, generate_relative_import


class GlobalControllerGenerator(FileGeneratorBase):
    """
    Generate the root/controller.ts file for all global definitions.
    - Models
    - Enums
    - Controllers

    """

    def __init__(
        self,
        *,
        controller_wrappers: list[ControllerWrapper],
        managed_path: ManagedViewPath,
    ):
        super().__init__(managed_path=managed_path)
        self.controller_wrappers = controller_wrappers

    def script(self):
        # Recursively traverse our controller definitions, which themselves point to other
        # in-memory objects that should be converted like models and enums
        controllers = ControllerWrapper.get_all_embedded_controllers(
            self.controller_wrappers
        )
        embedded_types = ControllerWrapper.get_all_embedded_types(
            controllers, include_superclasses=True
        )

        # Resolve the MRO ordering for all the interfaces, since they'll be defined
        # in one file
        controller_sorted = self._build_controller_graph(controllers)
        model_enum_sorted = self._build_model_enum_graph(
            embedded_types.models, embedded_types.enums
        )

        yield CodeBlock(
            "/*",
            " * Models + Enums",
            " */",
        )

        # Models and enums will be used by the action signatures contained
        # in the controllers
        for item in model_enum_sorted:
            if isinstance(item, ModelWrapper):
                yield CodeBlock(ModelInterface.from_model(item).to_js())
            elif isinstance(item, EnumWrapper):
                yield CodeBlock(EnumInterface.from_enum(item).to_js())
            else:
                raise ValueError(f"Unsupported item type: {item}")

        yield CodeBlock(
            "/*",
            " * Exceptions",
            " */",
        )

        # Exceptions don't have other dependencies so they can be added in any order
        for exception in embedded_types.exceptions:
            yield CodeBlock(ExceptionInterface.from_exception(exception).to_js())

        yield CodeBlock(
            "/*",
            " * View Controllers",
            " */",
        )

        # Then process controllers in dependency order
        for controller in controller_sorted:
            yield CodeBlock(ControllerInterface.from_controller(controller).to_js())

    def _build_model_enum_graph(
        self, models: list[ModelWrapper], enums: list[EnumWrapper]
    ):
        """Build dependency graph for models and enums"""
        # Build id-based graph
        graph: dict[int, set[int]] = {}
        id_to_obj: dict[int, ModelWrapper | EnumWrapper] = {}

        # Initialize graph entries for all models and enums
        for model in models:
            graph[id(model)] = set()
            id_to_obj[id(model)] = model

        for enum in enums:
            graph[id(enum)] = set()
            id_to_obj[id(enum)] = enum

        # Add model superclass dependencies
        for model in models:
            graph[id(model)].update(id(superclass) for superclass in model.superclasses)

            # Add field dependencies
            for field in model.value_models:
                if isinstance(field.value, (ModelWrapper, EnumWrapper)):
                    graph[id(model)].add(id(field.value))

        # Convert graph to use actual objects for TopologicalSorter
        sorted_ids = TopologicalSorter(graph).static_order()
        return [id_to_obj[node_id] for node_id in sorted_ids]

    def _build_controller_graph(
        self, controllers: list[ControllerWrapper]
    ) -> list[ControllerWrapper]:
        """Build dependency graph for controllers"""
        # Build id-based graph
        graph: dict[int, set[int]] = {}
        id_to_obj: dict[int, ControllerWrapper] = {}

        # Initialize graph entries for all controllers
        for controller in controllers:
            graph[id(controller)] = set()
            id_to_obj[id(controller)] = controller

        # Add controller superclass dependencies
        for controller in controllers:
            graph[id(controller)].update(
                id(superclass) for superclass in controller.superclasses
            )

        # Convert graph to use actual objects for TopologicalSorter
        sorted_ids = TopologicalSorter(graph).static_order()
        return [id_to_obj[node_id] for node_id in sorted_ids]


class GlobalLinkGenerator(FileGeneratorBase):
    def __init__(
        self,
        *,
        parsed_controllers: list[ParsedController],
        managed_path: ManagedViewPath,
    ):
        super().__init__(managed_path=managed_path)
        self.parsed_controllers = parsed_controllers

    def script(self):
        """Generate global link aggregator"""
        imports: list[str] = []
        link_setters: dict[str, Any] = {}

        for parsed_controller in self.parsed_controllers:
            if parsed_controller.is_layout:
                continue

            controller_dir = parsed_controller.view_path.get_managed_code_dir()
            controller_implementation_path = generate_relative_import(
                self.managed_path, controller_dir / "links.ts"
            )

            # Add import and setter for this controller
            local_name = f"{parsed_controller.wrapper.name.global_name}GetLinks"
            imports.append(
                f"import {{ getLink as {local_name} }} from '{controller_implementation_path}';"
            )
            link_setters[
                # @pierce: 12-11-2024: Mirror the lowercase camelcase convention of previous versions
                TSLiteral(
                    camelize(
                        parsed_controller.wrapper.controller.__name__,
                        uppercase_first_letter=False,
                    )
                )
            ] = TSLiteral(local_name)

        link_generator = python_payload_to_typescript(link_setters)

        yield CodeBlock(*imports)
        yield CodeBlock(f"const linkGenerator = {link_generator};")
        yield CodeBlock("export default linkGenerator;")



from contextlib import asynccontextmanager
from inspect import Parameter, signature
from pathlib import Path
from unittest.mock import patch

import pytest
from fastapi import APIRouter, FastAPI, Request, status
from fastapi.exceptions import RequestValidationError as RequestValidationErrorRaw
from fastapi.responses import RedirectResponse
from fastapi.routing import APIRoute
from fastapi.testclient import TestClient
from pydantic import BaseModel, ValidationError

from mountaineer.app import AppController, ControllerDefinition
from mountaineer.config import ConfigBase
from mountaineer.controller import ControllerBase
from mountaineer.controller_layout import LayoutControllerBase
from mountaineer.exceptions import (
    APIException,
    RequestValidationError,
    RequestValidationFailure,
)
from mountaineer.render import Metadata, RenderBase


def test_requires_render_return_value():
    """
    The AppController is in charge of validating our render return value. Since renders are not
    decorated, the best place to validate is during a mount.

    """

    class TestControllerWithoutRenderMarkup(ControllerBase):
        url = "/"
        view_path = "/page.tsx"

        def render(self):
            return None

    class TestControllerWithRenderMarkup(ControllerBase):
        url = "/"
        view_path = "/page.tsx"

        def render(self) -> None:
            return None

    app = AppController(view_root=Path(""))
    with pytest.raises(ValueError, match="must have a return type annotation"):
        app.register(TestControllerWithoutRenderMarkup())

    app.register(TestControllerWithRenderMarkup())


def test_validates_layouts_exclude_urls():
    """
    The app controller should reject the registration of layouts that specify
    a url.

    """

    class TestLayoutController(LayoutControllerBase):
        # Not allowed, but might typehint correctly because the ControllerBase
        # superclass supports it.
        url = "/layout_url"
        view_path = "/test.tsx"

        async def render(self) -> None:
            pass

    app_controller = AppController(view_root=Path(""))
    with pytest.raises(ValueError, match="are not directly mountable to the router"):
        app_controller.register(TestLayoutController())


def test_format_exception_model():
    class ExampleException(APIException):
        status_code = 401
        value: str

    app = AppController(view_root=Path(""))
    formatted_exception = app._format_exception_model(ExampleException)

    assert formatted_exception.status_code == 401
    assert formatted_exception.schema_name == "ExampleException"
    assert (
        formatted_exception.schema_name_long
        == "mountaineer.__tests__.test_app.ExampleException"
    )
    assert set(formatted_exception.schema_value["required"]) == {
        "value",
        # Inherited from the superclass
        "status_code",
        "detail",
        "headers",
    }


def test_view_root_from_config(tmp_path: Path):
    class MockConfig(ConfigBase):
        PACKAGE: str | None = "test_webapp"

    # Simulate a package with a views directory
    (tmp_path / "views").mkdir()

    with patch("mountaineer.app.resolve_package_path") as mock_resolve_package_path:
        mock_resolve_package_path.return_value = tmp_path

        app = AppController(config=MockConfig())
        assert app._view_root == tmp_path / "views"

        assert mock_resolve_package_path.call_count == 1
        assert mock_resolve_package_path.call_args[0] == ("test_webapp",)


def test_passthrough_fastapi_args():
    did_run_lifespan = False

    @asynccontextmanager
    async def app_lifespan(app: FastAPI):
        nonlocal did_run_lifespan
        did_run_lifespan = True
        yield

    app = AppController(view_root=Path(""), fastapi_args=dict(lifespan=app_lifespan))

    with TestClient(app.app):
        assert did_run_lifespan


def test_unique_controller_names():
    def make_controller(unique_url: str):
        class ExampleController(ControllerBase):
            url = unique_url
            view_path = unique_url

            def render(self) -> None:
                pass

        return ExampleController

    app = AppController(view_root=Path(""))
    app.register(make_controller("/example")())

    with pytest.raises(ValueError, match="already registered"):
        app.register(make_controller("/example2")())


class TargetController(ControllerBase):
    url = "/target"

    async def render(self) -> None:
        pass


class ReferenceController(ControllerBase):
    url = "/reference"

    async def render(self) -> None:
        pass


def test_merge_render_signatures():
    def target_fn(a: int, b: int):
        pass

    # Partial overlap with (a) and inclusion of a new variable
    def reference_fn(a: int, c: int):
        pass

    app = AppController(view_root=Path(""))

    target_definition = ControllerDefinition(
        controller=TargetController(),
        router=APIRouter(),
        view_route=target_fn,
        url_prefix="/target_prefix",
        render_router=APIRouter(),
    )
    reference_definition = ControllerDefinition(
        controller=ReferenceController(),
        router=APIRouter(),
        view_route=reference_fn,
        url_prefix="/reference_prefix",
        render_router=APIRouter(),
    )

    initial_routes = [
        route.path for route in app.app.routes if isinstance(route, APIRoute)
    ]
    assert initial_routes == []

    app.merge_render_signatures(
        target_definition, reference_controller=reference_definition
    )

    assert list(signature(target_definition.view_route).parameters.values()) == [
        Parameter("a", Parameter.POSITIONAL_OR_KEYWORD, annotation=int),
        Parameter("b", Parameter.POSITIONAL_OR_KEYWORD, annotation=int),
        # Items only in the reference function should be included as kwargs
        Parameter("c", Parameter.KEYWORD_ONLY, annotation=int, default=Parameter.empty),
    ]

    # After the merging the signature should be updated, and the app controller should
    # have a new endpoint (since the merging must re-mount)
    final_routes = [
        route.path for route in app.app.routes if isinstance(route, APIRoute)
    ]
    assert final_routes == ["/target"]


def test_merge_render_signatures_conflicting_types():
    """
    If the two functions share a parameter, it must be typehinted with the
    same type in both functions.

    """

    def target_fn(a: int, b: int):
        pass

    # Partial overlap with (a) and inclusion of a new variable
    def reference_fn(a: str, c: int):
        pass

    app = AppController(view_root=Path(""))

    target_definition = ControllerDefinition(
        controller=TargetController(),
        router=APIRouter(),
        view_route=target_fn,
        url_prefix="/target_prefix",
        render_router=APIRouter(),
    )
    reference_definition = ControllerDefinition(
        controller=ReferenceController(),
        router=APIRouter(),
        view_route=reference_fn,
        url_prefix="/reference_prefix",
        render_router=APIRouter(),
    )

    with pytest.raises(TypeError, match="Conflicting types"):
        app.merge_render_signatures(
            target_definition, reference_controller=reference_definition
        )


def test_get_value_mask_for_signature():
    def target_fn(a: int, b: str):
        pass

    values = {
        "a": 1,
        "b": "test",
        "c": "other",
    }

    app = AppController(view_root=Path(""))
    assert app._get_value_mask_for_signature(
        signature(target_fn),
        values,
    ) == {
        "a": 1,
        "b": "test",
    }


class RedirectRender(RenderBase):
    pass


class RedirectController(ControllerBase):
    url = "/redirect"
    view_path = "/test.tsx"

    async def render(self) -> RedirectRender:
        return RedirectRender(
            metadata=Metadata(
                explicit_response=RedirectResponse(
                    status_code=status.HTTP_307_TEMPORARY_REDIRECT, url="/"
                )
            )
        )


def test_explicit_response_metadata():
    app = AppController(view_root=Path(""))
    app.register(RedirectController())

    with TestClient(app.app) as client:
        response = client.get("/redirect", follow_redirects=False)
        assert response.status_code == status.HTTP_307_TEMPORARY_REDIRECT
        assert response.headers["location"] == "/"


@pytest.mark.asyncio
async def test_parse_validation_exception():
    """
    Test that FastAPI validation errors are correctly parsed into our RequestValidationError format.
    """

    class TestModel(BaseModel):
        age: int

    app_controller = AppController(view_root=Path(""))

    # Create a test request with invalid data
    request = Request(
        scope={
            "type": "http",
            "method": "POST",
            "path": "/",
            "headers": [],
        }
    )

    # Create a validation error by trying to validate invalid data
    raw_error: RequestValidationErrorRaw | None = None
    try:
        TestModel.model_validate({"age": "not_a_number"})
    except ValidationError as e:
        raw_error = RequestValidationErrorRaw(errors=e.errors())

    # Test the parsing
    assert raw_error
    with pytest.raises(RequestValidationError) as exc_info:
        await app_controller._parse_validation_exception(request, raw_error)

    exception = exc_info.value
    assert len(exception.internal_model.errors) == 1  # type: ignore
    error = exception.internal_model.errors[0]  # type: ignore
    assert isinstance(error, RequestValidationFailure)

    # Verify the error is parsed correctly
    assert error.error_type == "int_parsing"
    assert error.location == ["age"]
    assert "input should be a valid integer" in error.message.lower()
    assert error.value_input == "not_a_number"



from mountaineer.__tests__.development.conftest import AppPackageType
from mountaineer.development.packages import package_path_to_module


def test_package_path_to_module(app_package: AppPackageType):
    package_name, package_path, _ = app_package
    file_path = package_path / "test_controller.py"
    module_name = package_path_to_module(package_name, file_path)

    assert module_name == f"{package_name}.test_controller"



from enum import Enum, auto
from typing import Union

import pytest

from mountaineer.__tests__.client_builder.interface_builders.common import (
    create_enum_wrapper,
)
from mountaineer.client_builder.interface_builders.enum import EnumInterface


class TestBasicEnumGeneration:
    def test_string_enum(self):
        class StringEnum(Enum):
            ALPHA = "alpha"
            BETA = "beta"
            GAMMA = "gamma"

        interface = EnumInterface.from_enum(create_enum_wrapper(StringEnum))
        ts_code = interface.to_js()

        assert "export enum StringEnum" in ts_code
        assert "ALPHA = " in ts_code
        assert "'alpha'" in ts_code
        assert ts_code.count(",") == 2  # Two commas for three values

    def test_number_enum(self):
        class NumberEnum(Enum):
            ONE = 1
            TWO = 2
            THREE = 3

        interface = EnumInterface.from_enum(create_enum_wrapper(NumberEnum))
        ts_code = interface.to_js()

        assert "ONE = 1" in ts_code
        assert "TWO = 2" in ts_code
        assert "THREE = 3" in ts_code

    def test_mixed_type_enum(self):
        class MixedEnum(Enum):
            STRING = "value"
            NUMBER = 42
            BOOLEAN = True
            NULL = None

        interface = EnumInterface.from_enum(create_enum_wrapper(MixedEnum))
        ts_code = interface.to_js()

        assert "'value'" in ts_code
        assert "42" in ts_code
        assert "true" in ts_code
        assert "null" in ts_code


class TestEnumFormatting:
    def test_export_statement(self):
        class ExampleEnum(Enum):
            A = "a"
            B = "b"

        interface = EnumInterface.from_enum(create_enum_wrapper(ExampleEnum))

        # Test with export
        assert interface.to_js().startswith("export enum")

        # Test without export
        interface.include_export = False
        assert interface.to_js().startswith("enum")
        assert "export" not in interface.to_js()

    def test_enum_structure(self):
        class ExampleEnum(Enum):
            A = "a"
            B = "b"
            C = "c"

        interface = EnumInterface.from_enum(create_enum_wrapper(ExampleEnum))
        ts_code = interface.to_js()

        assert ts_code.count("{") == 1
        assert ts_code.count("}") == 1
        assert ts_code.count(",") == len(ExampleEnum) - 1

    @pytest.mark.parametrize(
        "value, expected",
        [
            ("string", "'string'"),
            (42, "42"),
            (True, "true"),
            (None, "null"),
        ],
    )
    def test_value_formatting(self, value: str, expected: Union[str, int, bool, None]):
        class ValueEnum(Enum):
            TEST = value

        interface = EnumInterface.from_enum(create_enum_wrapper(ValueEnum))
        ts_code = interface.to_js()

        assert f"TEST = {expected}" in ts_code


class TestComplexCases:
    def test_auto_enum(self):
        class AutoEnum(Enum):
            FIRST = auto()
            SECOND = auto()
            THIRD = auto()

        interface = EnumInterface.from_enum(create_enum_wrapper(AutoEnum))
        ts_code = interface.to_js()

        assert "FIRST = 1" in ts_code
        assert "SECOND = 2" in ts_code
        assert "THIRD = 3" in ts_code

    def test_duplicate_values(self):
        class DuplicateValueEnum(Enum):
            A = "value"
            B = "value"
            C = "value"

        interface = EnumInterface.from_enum(create_enum_wrapper(DuplicateValueEnum))
        ts_code = interface.to_js()

        assert "A = " in ts_code
        assert "B = " in ts_code
        assert "C = " in ts_code
        assert ts_code.count("'value'") == 3


class TestEdgeCases:
    def test_single_member_enum(self):
        class SingleEnum(Enum):
            ONLY = "only"

        interface = EnumInterface.from_enum(create_enum_wrapper(SingleEnum))
        ts_code = interface.to_js()

        assert "ONLY = " in ts_code
        assert ts_code.count(",") == 0

    def test_special_characters(self):
        class SpecialCharEnum(Enum):
            DASH_VALUE = "dash-value"
            UNDERSCORE_VALUE = "underscore_value"
            SPACE_VALUE = "space value"

        interface = EnumInterface.from_enum(create_enum_wrapper(SpecialCharEnum))
        ts_code = interface.to_js()

        assert "DASH_VALUE = 'dash-value'" in ts_code
        assert "UNDERSCORE_VALUE = 'underscore_value'" in ts_code
        assert "SPACE_VALUE = 'space value'" in ts_code

    def test_empty_enum(self):
        class EmptyEnum(Enum):
            pass

        interface = EnumInterface.from_enum(create_enum_wrapper(EmptyEnum))
        ts_code = interface.to_js()

        assert "enum EmptyEnum {" in ts_code
        assert ts_code.strip().endswith("}")



from dataclasses import dataclass
from enum import Enum


class PackageManager(str, Enum):
    UV = "uv"
    POETRY = "poetry"
    VENV = "venv"


@dataclass
class EditorDescription:
    name: str
    path: str | None


class EditorType(Enum):
    VSCODE = EditorDescription(name="vscode", path="vscode")
    VIM = EditorDescription(name="vim", path="vim")
    ZED = EditorDescription(name="zed", path=None)

    @classmethod
    def from_name(cls, name: str) -> "EditorType":
        for editor in cls:
            if editor.value.name == name:
                return editor
        raise ValueError(f"Invalid editor type: {name}")



from datetime import date, datetime, time
from enum import Enum
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)
from uuid import UUID

import pytest
from fastapi import UploadFile
from pydantic import BaseModel

from mountaineer.client_builder.interface_builders.base import InterfaceBase
from mountaineer.client_builder.parser import (
    EnumWrapper,
    ModelWrapper,
    SelfReference,
    WrapperName,
)
from mountaineer.client_builder.types import (
    DictOf,
    ListOf,
    LiteralOf,
    Or,
    SetOf,
    TupleOf,
    TypeDefinition,
)


# Test Models and Types
class Status(Enum):
    ACTIVE = "active"
    INACTIVE = "inactive"
    PENDING = "pending"


class SimpleModel(BaseModel):
    name: str
    count: int
    active: bool


class NestedModel(BaseModel):
    status: Status
    data: SimpleModel
    optional: Optional[SimpleModel]


class ComplexTypes(BaseModel):
    list_field: List[str]
    dict_field: Dict[str, int]
    set_field: Set[float]
    tuple_field: Tuple[str, int]
    union_field: Union[str, int]
    literal_field: Literal["a", "b", "c"]


# Test Implementation
T = TypeVar("T")


class TypeConverter(InterfaceBase):
    """Helper class for testing the abstract base model"""

    @classmethod
    def convert(cls, value: Any) -> str:
        return cast(str, cls._get_annotated_value(value))


# Test Fixtures
@pytest.fixture
def model_wrapper() -> ModelWrapper:
    return ModelWrapper(
        name=WrapperName("SimpleModel"),
        module_name="test_module",
        model=SimpleModel,
        isolated_model=SimpleModel,
        superclasses=[],
        value_models=[],
    )


@pytest.fixture
def enum_wrapper() -> EnumWrapper:
    return EnumWrapper(
        name=WrapperName("Status"),
        module_name="test_module",
        enum=Status,
    )


@pytest.fixture
def self_reference() -> SelfReference:
    return SelfReference(
        name="CircularModel",
        model=SimpleModel,
    )


class TestPrimitiveTypeMapping:
    @pytest.mark.parametrize(
        "py_type,ts_type",
        [
            (str, "string"),
            (int, "number"),
            (float, "number"),
            (bool, "boolean"),
            (datetime, "string"),
            (date, "string"),
            (time, "string"),
            (UUID, "string"),
            (UploadFile, "Blob"),
            (None, "null"),
            (Any, "any"),
        ],
    )
    def test_primitive_type_mapping(self, py_type: Type[Any], ts_type: str) -> None:
        result: str = TypeConverter.convert(py_type)
        assert result == ts_type

    def test_none_type_variations(self) -> None:
        assert TypeConverter.convert(None) == "null"
        assert TypeConverter.convert(type(None)) == "null"

    def test_any_type_fallback(self) -> None:
        class CustomType:
            pass

        result: str = TypeConverter.convert(CustomType)
        assert result == "any"


class TestComplexTypeHandling:
    def test_list_conversion(self) -> None:
        type_def: ListOf = ListOf(str)
        result: str = TypeConverter.convert(type_def)
        assert result == "Array<string>"

    def test_nested_list_conversion(self) -> None:
        type_def: ListOf = ListOf(ListOf(str))
        result: str = TypeConverter.convert(type_def)
        assert result == "Array<Array<string>>"

    def test_dict_conversion(self) -> None:
        type_def: DictOf = DictOf(str, int)
        result: str = TypeConverter.convert(type_def)
        assert result == "Record<string, number>"

    def test_set_conversion(self) -> None:
        type_def: SetOf = SetOf(float)
        result: str = TypeConverter.convert(type_def)
        assert result == "Set<number>"

    def test_tuple_conversion(self) -> None:
        type_def = TupleOf(str, int, bool)
        result: str = TypeConverter.convert(type_def)
        assert result == "[string,number,boolean]"

    def test_union_conversion(self) -> None:
        type_def = Or(str, int, bool)
        result: str = TypeConverter.convert(type_def)
        assert result == "string | number | boolean"

    def test_literal_conversion(self) -> None:
        type_def = LiteralOf("active", "inactive")
        result: str = TypeConverter.convert(type_def)
        assert result == '"active" | "inactive"'

    @pytest.mark.parametrize(
        "type_def,expected",
        [
            (DictOf(str, ListOf(int)), "Record<string, Array<number>>"),
            (ListOf(DictOf(str, bool)), "Array<Record<string, boolean>>"),
            (
                Or(ListOf(str), DictOf(str, int)),
                "Array<string> | Record<string, number>",
            ),
        ],
    )
    def test_nested_complex_types(
        self, type_def: TypeDefinition, expected: str
    ) -> None:
        result: str = TypeConverter.convert(type_def)
        assert result == expected


class TestModelHandling:
    def test_model_wrapper_conversion(self, model_wrapper: ModelWrapper) -> None:
        result: str = TypeConverter.convert(model_wrapper)
        assert result == "SimpleModel"

    def test_enum_wrapper_conversion(self, enum_wrapper: EnumWrapper) -> None:
        result: str = TypeConverter.convert(enum_wrapper)
        assert result == "Status"

    def test_self_reference_conversion(self, self_reference: SelfReference) -> None:
        result: str = TypeConverter.convert(self_reference)
        assert result == "CircularModel"

    def test_nested_model_references(self, model_wrapper: ModelWrapper) -> None:
        type_def: ListOf = ListOf(model_wrapper)
        result: str = TypeConverter.convert(type_def)
        assert result == "Array<SimpleModel>"


class TestComplexScenarios:
    def test_deeply_nested_structure(self) -> None:
        # Create a deeply nested structure
        deep_type = ListOf(
            DictOf(str, Or(ListOf(TupleOf(str, int)), DictOf(str, SetOf(float))))
        )
        result = TypeConverter.convert(deep_type)
        expected = "Array<Record<string, Array<[string,number]> | Record<string, Set<number>>>>"
        assert result == expected

    def test_mixed_model_and_primitive_types(
        self, model_wrapper: ModelWrapper, enum_wrapper: EnumWrapper
    ) -> None:
        type_def = Or(model_wrapper, ListOf(enum_wrapper), DictOf(str, int))
        result: str = TypeConverter.convert(type_def)
        assert "SimpleModel | Array<Status> | Record<string, number>" == result

    def test_complex_union_types(self) -> None:
        # Test union with various nested types
        type_def = Or(
            ListOf(str),
            DictOf(str, bool),
            SetOf(int),
            TupleOf(str, int),
            LiteralOf("a", "b"),
        )
        result: str = TypeConverter.convert(type_def)
        expected: str = 'Array<string> | Record<string, boolean> | Set<number> | [string,number] | "a" | "b"'
        assert result == expected

    @pytest.mark.parametrize(
        "type_def,expected",
        [
            (
                DictOf(LiteralOf("id", "name"), Or(str, int)),
                'Record<"id" | "name", string | number>',
            ),
            (
                ListOf(TupleOf(LiteralOf("GET", "POST"), str)),
                'Array<["GET" | "POST",string]>',
            ),
            (
                SetOf(Or(LiteralOf(1, 2), LiteralOf("a", "b"))),
                'Set<1 | 2 | "a" | "b">',
            ),
        ],
    )
    def test_complex_literal_combinations(
        self, type_def: TypeDefinition, expected: str
    ) -> None:
        result: str = TypeConverter.convert(type_def)
        assert result == expected



from textwrap import dedent

import pytest

from mountaineer.client_builder.file_generators.base import CodeBlock


class TestCodeBlock:
    def test_initialization(self):
        simple_code_block = CodeBlock("line1", "line2")
        assert simple_code_block.lines == ("line1", "line2")

    @pytest.mark.parametrize(
        "input_str, expected",
        [
            # Empty string cases
            ("", ""),
            # Single line cases
            ("hello", "hello"),
            ("    hello", "    hello"),
            ("\thello", "\thello"),
            # Basic multi-line cases - adding to subsequent indentation
            (
                "    first\n        second\n            third",
                "    first\n            second\n                third",
            ),
            (
                "\tfirst\n\t\tsecond\n\t\t\tthird",
                "\tfirst\n\t\t\tsecond\n\t\t\t\tthird",
            ),
            # Cases with trailing newlines
            ("    first\n        second\n", "    first\n            second\n"),
            # Cases with empty lines
            ("    first\n\n        third", "    first\n\n            third"),
            ("    first\n  \n        third", "    first\n  \n            third"),
            # Cases with mixed indentation - adding first line indent
            (
                "    first\n  second\n      third",
                "    first\n      second\n          third",
            ),
            # Cases with tabs and spaces
            ("\tfirst\n    second\n\t\tthird", "\tfirst\n\t    second\n\t\t\tthird"),
            # Complex multi-line string cases
            (
                dedent(
                    """
            def example():
                x = '''
                multi
                line
                string'''
            """
                ).strip(),
                "def example():\n    x = '''\n    multi\n    line\n    string'''",
            ),
        ],
    )
    def test_indent_variations(self, input_str, expected):
        result = CodeBlock.indent(input_str)
        assert result == expected

    def test_real_world_template_strings(self):
        # Test with f-strings and template variables
        var = "hello\n    world"  # Note: intentional indentation in var
        input_str = f"    print({var})"
        expected = "    print(hello\n        world)"
        assert CodeBlock.indent(input_str) == expected

        # Test with triple-quoted strings
        var = """first
            second
                third"""
        input_str = f"    function({var})"
        expected = (
            "    function(first\n                second\n                    third)"
        )
        assert CodeBlock.indent(input_str) == expected

    def test_whitespace_preservation(self):
        # Test that trailing whitespace is preserved
        input_str = "    x = 1  # trailing spaces  \n        y = 2"
        expected = "    x = 1  # trailing spaces  \n            y = 2"
        assert CodeBlock.indent(input_str) == expected

        # Test with tabs and spaces mixed
        input_str = "\t    x = 1\n\t\t    y = 2"
        expected = "\t    x = 1\n\t    \t\t    y = 2"
        assert CodeBlock.indent(input_str) == expected

    def test_multiple_newlines(self):
        # Test handling of multiple consecutive newlines
        input_str = "    first\n\n\n        second"
        expected = "    first\n\n\n            second"
        assert CodeBlock.indent(input_str) == expected

    def test_unicode_handling(self):
        # Test with unicode characters
        input_str = "    🐍 = 'python'\n        © = 'copyright'"
        expected = "    🐍 = 'python'\n            © = 'copyright'"
        assert CodeBlock.indent(input_str) == expected



from hashlib import md5
from pathlib import Path
from shutil import move as shutil_move
from tempfile import mkdtemp

from mountaineer.app import AppController
from mountaineer.client_compiler.base import ClientBundleMetadata
from mountaineer.client_compiler.build_metadata import BuildMetadata
from mountaineer.client_compiler.exceptions import BuildProcessException
from mountaineer.console import CONSOLE
from mountaineer.io import gather_with_concurrency
from mountaineer.logging import LOGGER
from mountaineer.paths import ManagedViewPath


class ClientCompiler:
    """
    Main entrypoint for compiling client code into the artifacts that
    will be used in production.

    """

    def __init__(
        self,
        app: AppController,
    ):
        self.tmp_dir = Path(mkdtemp())

        self.update_controller(app)

    def update_controller(self, controller: AppController):
        self.app = controller
        self.view_root = controller._view_root

    async def run_builder_plugins(
        self,
        limit_paths: list[Path] | None = None,
        max_concurrency: int = 10,
    ):
        """
        :param limit_paths: If not provided, will sniff out all the files on disk. Note that
        this is a None equality check; a blank list is provided, we won't refresh any files.

        """
        if limit_paths is None:
            limit_paths = list(self._get_static_files())
            limit_paths += [
                self.view_root.get_controller_view_path(
                    controller_definition.controller
                )
                for controller_definition in self.app.controllers
            ]

        if not limit_paths:
            LOGGER.debug("No files to compile in builder plugins, skipping...")
            return

        # For now we do this every time to make sure we pick up on
        # new controllers that might have been added since this class
        # was created
        self._init_builders()

        for path in limit_paths:
            for builder in self.app.builders:
                builder.mark_file_dirty(path)

        results = await gather_with_concurrency(
            [builder.build_wrapper() for builder in self.app.builders],
            n=max_concurrency,
            catch_exceptions=True,
        )

        # This seems to stall our writer
        # LOGGER.debug(f"Plugin builders took {(monotonic_ns() - start) / 1e9}s")

        # Go through the exceptions, logging the build errors explicitly
        has_build_error = False
        final_exception: str = ""
        for result in results:
            if isinstance(result, Exception):
                has_build_error = True
                if isinstance(result, BuildProcessException):
                    CONSOLE.print(f"[bold red]Build error: {result}")
                final_exception += str(result)

        if has_build_error:
            raise BuildProcessException(final_exception)

        # Up until now builders have placed their results into a temporary
        # directory, we want to merge this with the project directory
        self._move_build_artifacts_into_project()

        # Now that we have prepared the static folder in its final form, we
        # should get the md5 hash of the content for our archive
        metadata = self._build_static_metadata()
        (self.view_root.get_managed_metadata_dir() / "metadata.json").write_text(
            metadata.model_dump_json()
        )

    def _init_builders(self):
        metadata = ClientBundleMetadata(
            live_reload_port=None,
            package_root_link=self.view_root.get_package_root_link(),
            tmp_dir=self.tmp_dir,
        )

        for builder in self.app.builders:
            builder.set_metadata(metadata)

        for builder in self.app.builders:
            for controller_definition in self.app.controllers:
                builder.register_controller(
                    controller_definition.controller,
                    self.view_root.get_controller_view_path(
                        controller_definition.controller
                    ),
                )

    def _get_static_files(self):
        ignore_directories = ["_ssr", "_static", "_server", "_metadata", "node_modules"]

        for view_root in self._get_all_root_views():
            for dir_path, _, filenames in view_root.walk():
                for filename in filenames:
                    if any(
                        [
                            directory in dir_path.parts
                            for directory in ignore_directories
                        ]
                    ):
                        continue
                    yield dir_path / filename

    def _get_all_root_views(self) -> list[ManagedViewPath]:
        """
        The self.view_root variable is the root of the current user project. We may have other
        "view roots" that store view for plugins.

        This function inspects the controller path definitions and collects all of the
        unique root view paths. The returned ManagedViewPaths are all copied and set to
        share the same package root as the user project.

        """
        # Find the view roots
        view_roots = {self.view_root.copy()}
        for controller_definition in self.app.controllers:
            view_path = controller_definition.controller.view_path
            if isinstance(view_path, ManagedViewPath):
                view_roots.add(view_path.get_root_link().copy())

        # All the view roots should have the same package root
        for view_root in view_roots:
            view_root.package_root_link = self.view_root.package_root_link

        return list(view_roots)

    def _move_build_artifacts_into_project(self):
        """
        Now that we build has completed, we can clear out the old files and replace it
        with the thus-far temporary files

        This cleans up old controllers in the case that they were deleted, and prevents
        outdated md5 content hashes from being served

        """
        # For now, just copy over the _tmp files into the main directory. Replace them
        # if they exist. This is a differential copy to support our build pipeline
        # that will only change progressively. It notably does not handle the case of
        # deleted files.
        for tmp_path, final_path in [
            (
                self.view_root.get_managed_static_dir(tmp_build=True),
                self.view_root.get_managed_static_dir(),
            ),
            # (
            #    self.view_root.get_managed_ssr_dir(tmp_build=True),
            #    self.view_root.get_managed_ssr_dir(),
            # ),
        ]:
            # Only remove the final path if the matching tmp path build exists
            # This creates a merged build where we maintain the base code created
            # on startup and the possibly incremental builds.
            # We do the deletion up-front so we don't delete our actual files mid-copy
            # in case we have multiple with the same prefix (like .js and .map)
            for tmp_file in tmp_path.glob("*"):
                # Strip any md5 hashes from the filename and delete
                # all the items matching the same name
                base_filename = tmp_file.name.rsplit("-")[0].rsplit(".")[0]
                for old_file in final_path.glob(f"{base_filename}*"):
                    old_file.unlink()

            # We've cleared old versions, now freshly copy the files
            for tmp_file in tmp_path.glob("*"):
                final_file = final_path / tmp_file.name
                shutil_move(tmp_file, final_file)

    def _build_static_metadata(self):
        static_artifact_shas: dict[str, str] = {}
        static_dir = self.view_root.get_managed_static_dir()

        for base_path, dirs, files in static_dir.walk():
            for file in files:
                static_path = base_path / file
                file_contents = static_path.read_bytes()
                static_artifact_shas[str(static_path.relative_to(static_dir))] = md5(
                    file_contents
                ).hexdigest()

        return BuildMetadata(static_artifact_shas=static_artifact_shas)



import traceback
from contextlib import contextmanager
from hashlib import md5
from multiprocessing import get_start_method, set_start_method
from os import getenv
from time import time
from typing import Any, Callable, Coroutine

from firehot import isolate_imports
from inflection import underscore
from rich.traceback import install as rich_traceback_install

from mountaineer import mountaineer as mountaineer_rs  # type: ignore
from mountaineer.console import CONSOLE
from mountaineer.constants import KNOWN_JS_EXTENSIONS
from mountaineer.development.isolation import IsolatedAppContext
from mountaineer.development.manager import (
    FileChangesState,
    IsolatedContext,
    WebserverConfig,
    rebuild_frontend,
    restart_backend,
)
from mountaineer.development.messages import (
    SuccessResponse,
)
from mountaineer.development.messages_broker import (
    AsyncMessageBroker,
    BrokerExecutionError,
)
from mountaineer.development.packages import (
    find_packages_with_prefix,
)
from mountaineer.development.watch import (
    CallbackDefinition,
    CallbackMetadata,
    CallbackType,
    PackageWatchdog,
)
from mountaineer.development.watch_server import WatcherWebservice
from mountaineer.io import async_to_sync
from mountaineer.logging import LOGGER
from mountaineer.ssr import find_tsconfig
from mountaineer.static import get_static_path


@async_to_sync
async def handle_watch(
    *,
    package: str,
    webcontroller: str,
    subscribe_to_mountaineer: bool = False,
):
    """
    Watch the file directory and rebuild auto-generated files. This only
    creates the frontend files necessary to get server-side typehints. It
    doesn't build the package for production use.

    :param package: Ex. "ci_webapp"
    :param webcontroller: Ex. "ci_webapp.app:controller"
    :param subscribe_to_mountaineer:
        If True, will subscribe the local build server to changes in
        the `mountaineer` package. This is useful when doing concurrent
        development in `mountaineer` and a client package. Rarely
        used otherwise.

    """
    update_multiprocessing_settings()
    rich_traceback_install()

    file_changes_state = FileChangesState()
    first_run: bool = True

    async with AsyncMessageBroker.start_server() as (broker, config):
        isolated_context = IsolatedContext(
            webcontroller=webcontroller,
            webserver_config=None,
            message_config=config,
        )

        with get_mountaineer_isolated_env(package) as environment:
            CONSOLE.print("[bold blue]Development manager started")

            async def handle_file_changes(metadata: CallbackMetadata):
                try:
                    LOGGER.debug(f"Handling file changes: {metadata}")
                    nonlocal first_run
                    nonlocal file_changes_state

                    # First collect all the files that need updating
                    for event in metadata.events:
                        if event.path.suffix in KNOWN_JS_EXTENSIONS:
                            file_changes_state.pending_js.add(event.path)
                        elif event.path.suffix == ".py":
                            file_changes_state.pending_python.add(event.path)

                    if not first_run and not (
                        file_changes_state.pending_js
                        or file_changes_state.pending_python
                    ):
                        return

                    try:
                        if file_changes_state.pending_python or first_run:
                            await restart_backend(
                                environment,
                                broker,
                                file_changes_state,
                                isolated_context,
                            )

                        if file_changes_state.pending_js or first_run:
                            await rebuild_frontend(
                                broker,
                                file_changes_state,
                            )
                    except BrokerExecutionError as e:
                        CONSOLE.print(f"[red]Error: {e.error}\n\n{e.traceback}")
                        return

                    # If we've succeeded, we should clear out the pending
                    # files so we don't rebuild them again
                    file_changes_state.pending_js.clear()
                    file_changes_state.pending_python.clear()

                    first_run = False

                except Exception as e:
                    # Otherwise silently caught by our watchfiles command
                    CONSOLE.print(f"[red]Error: {e}")
                    CONSOLE.print(traceback.format_exc())
                    raise e

            watchdog = build_common_watchdog(
                package,
                handle_file_changes,
                subscribe_to_mountaineer=subscribe_to_mountaineer,
            )
            await watchdog.start_watching()

    CONSOLE.print("[green]Shutdown complete")


@async_to_sync
async def handle_runserver(
    *,
    package: str,
    webservice: str,
    webcontroller: str,
    host: str = "127.0.0.1",
    port: int,
    hotreload_host: str | None = None,
    hotreload_port: int | None = None,
    subscribe_to_mountaineer: bool = False,
):
    """
    Start a local development server. This will hot-reload your browser any time
    your frontend or backend code changes.

    :param package: Ex. "ci_webapp"
    :param webservice: Ex. "ci_webapp.app:app"
    :param webcontroller: Ex. "ci_webapp.app:controller"
    :param port: Desired port for the webapp while running locally
    :param subscribe_to_mountaineer: See `handle_watch` for more details.

    """
    update_multiprocessing_settings()
    rich_traceback_install()

    watcher_webservice = WatcherWebservice(
        webservice_host=hotreload_host or host, webservice_port=hotreload_port
    )
    await watcher_webservice.start()

    file_changes_state = FileChangesState()

    # Nonlocal vars for shutdown context
    watchdog: PackageWatchdog
    first_run: bool = True

    async with AsyncMessageBroker.start_server() as (broker, config):
        isolated_context = IsolatedContext(
            webcontroller=webcontroller,
            webserver_config=WebserverConfig(
                host=host,
                port=port,
                live_reload_port=watcher_webservice.port,
            ),
            message_config=config,
        )

        with get_mountaineer_isolated_env(package) as environment:
            CONSOLE.print("[bold blue]Development manager started")

            async def handle_file_changes(metadata: CallbackMetadata):
                try:
                    LOGGER.debug(f"Handling file changes: {metadata}")
                    nonlocal first_run
                    nonlocal file_changes_state

                    # First collect all the files that need updating
                    for event in metadata.events:
                        if event.path.suffix in KNOWN_JS_EXTENSIONS:
                            file_changes_state.pending_js.add(event.path)
                        elif event.path.suffix == ".py":
                            file_changes_state.pending_python.add(event.path)

                    if not first_run and not (
                        file_changes_state.pending_js
                        or file_changes_state.pending_python
                    ):
                        return

                    try:
                        if file_changes_state.pending_python or first_run:
                            await restart_backend(
                                environment,
                                broker,
                                file_changes_state,
                                isolated_context,
                            )

                        if file_changes_state.pending_js or first_run:
                            await rebuild_frontend(
                                broker,
                                file_changes_state,
                            )
                    except BrokerExecutionError as e:
                        CONSOLE.print(f"[red]Error: {e.error}\n\n{e.traceback}")
                        return

                    # If we've succeeded, we should clear out the pending
                    # files so we don't rebuild them again
                    file_changes_state.pending_js.clear()
                    file_changes_state.pending_python.clear()

                    # Ping the watcher webservice to let it know we've updated
                    await watcher_webservice.broadcast_listeners()

                    first_run = False

                except Exception as e:
                    # Otherwise silently caught by our watchfiles command
                    CONSOLE.print(f"[red]Error: {e}")
                    CONSOLE.print(traceback.format_exc())
                    raise e

            watchdog = build_common_watchdog(
                package,
                handle_file_changes,
                subscribe_to_mountaineer=subscribe_to_mountaineer,
            )
            await watchdog.start_watching()

    CONSOLE.print("[green]Shutdown complete")


@async_to_sync
async def handle_build(
    *,
    webcontroller: str,
    minify: bool = True,
):
    """
    Creates a production bundle of frontend files that is ready for service.

    Building your app will compile your TypeScript into the client-side bundle that will be downloaded
    by the browser. It also ahead-of-time generates the server code that will be run as part of [SSR](./ssr.md).
    You'll want to do it before deploying your application into production - but since a full build can take up
    to 10s, `handle_runserver` provides a better workflow for daily development.

    :param webcontroller: Ex. "ci_webapp.app:controller"
    :param minify: Minify the JS bundle, strip debug symbols

    """
    start = time()

    # Initialize the isolated context directly
    isolated_context = IsolatedAppContext.from_webcontroller(
        webcontroller=webcontroller,
        use_dev_exceptions=False,
    )

    # Initialize app state
    response = await isolated_context.initialize_app_state()
    if not isinstance(response, SuccessResponse):
        raise ValueError("Failed to initialize app state")

    # Type validation
    assert isolated_context.js_compiler is not None
    assert isolated_context.app_compiler is not None
    assert isolated_context.app_controller is not None

    # Build the frontend support bundle
    await isolated_context.js_compiler.build_use_server()
    await isolated_context.app_compiler.run_builder_plugins()

    # Get view paths for all controllers
    all_view_paths: list[list[str]] = []
    for controller_definition in isolated_context.app_controller.controllers:
        (
            _,
            direct_hierarchy,
        ) = isolated_context.app_controller._view_hierarchy_for_controller(
            controller_definition.controller
        )
        direct_hierarchy.reverse()
        all_view_paths.append([str(layout.path) for layout in direct_hierarchy])

    # Find tsconfig.json in the parent directories of the view paths
    tsconfig_path = find_tsconfig(all_view_paths)

    # Compile the final client bundle
    client_bundle_result = mountaineer_rs.compile_production_bundle(
        all_view_paths,
        str(isolated_context.app_controller._view_root / "node_modules"),
        "production",
        minify,
        str(get_static_path("live_reload.ts").resolve().absolute()),
        False,
        tsconfig_path,
    )

    static_output = isolated_context.app_controller._view_root.get_managed_static_dir()
    ssr_output = isolated_context.app_controller._view_root.get_managed_ssr_dir()

    # If we don't have the same number of entrypoints as controllers, something went wrong
    if len(client_bundle_result["entrypoints"]) != len(
        isolated_context.app_controller.controllers
    ):
        raise ValueError(
            f"Mismatch between number of controllers and number of entrypoints in the client bundle\n"
            f"Controllers: {len(isolated_context.app_controller.controllers)}\n"
            f"Entrypoints: {len(client_bundle_result['entrypoints'])}"
        )

    # Try to parse the format (entrypoint{}.js or entrypoint{}.js.map)
    for controller_definition, content, map_content in zip(
        isolated_context.app_controller.controllers,
        client_bundle_result["entrypoints"],
        client_bundle_result["entrypoint_maps"],
    ):
        script_root = underscore(controller_definition.controller.__class__.__name__)
        content_hash = md5(content.encode()).hexdigest()
        (static_output / f"{script_root}-{content_hash}.js").write_text(content)
        (static_output / f"{script_root}-{content_hash}.map.js").write_text(map_content)

    # Copy the other files 1:1 because they'll be referenced by name in the
    # entrypoints
    for path, content in client_bundle_result["supporting"].items():
        (static_output / path).write_text(content)

    # Now we go one-by-one to provide the SSR files, which will be consolidated
    # into a single runnable script for ease of use by the V8 engine
    result_scripts, _ = mountaineer_rs.compile_independent_bundles(
        all_view_paths,
        str(isolated_context.app_controller._view_root / "node_modules"),
        "production",
        0,
        str(get_static_path("live_reload.ts").resolve().absolute()),
        True,
        tsconfig_path,
    )

    # Write each script to disk
    for controller_definition, script in zip(
        isolated_context.app_controller.controllers, result_scripts
    ):
        script_root = underscore(controller_definition.controller.__class__.__name__)
        (ssr_output / f"{script_root}.js").write_text(script)

    LOGGER.info(f"Build completed in {(time() - start):.2f}s")


def update_multiprocessing_settings():
    """
    fork() is still the default on Linux, and can result in stalls with our asyncio
    event loops. For consistency and expected behavior, try to switch to spawn()
    if it's not already enabled.

    """
    if (spawn_method := get_start_method()) and spawn_method != "spawn":
        LOGGER.warning(
            f"The watch command should be run with the spawn start method, but it's currently set to {spawn_method}",
        )

        try:
            set_start_method("spawn", force=True)
            LOGGER.info("Start method is now set to 'spawn'.")
        except RuntimeError as e:
            # We might catch errors if set_start_method('spawn') is called
            # in an inappropriate context or after multiprocessing has started, which
            # is not allowed.
            LOGGER.error(f"Cannot change the start method after it has been used: {e}")


def build_common_watchdog(
    client_package: str,
    callback: Callable[[CallbackMetadata], Coroutine[Any, Any, None]],
    subscribe_to_mountaineer: bool,
):
    """
    Useful creation class to build a watchdog the common client class
    and our internal package.

    :param subscribe_to_mountaineer: If True, we'll also subscribe to the mountaineer package
    changes in the local environment. This is helpful for local development of the core
    package concurrent with a downstream client application.

    """
    dependent_packages: list[str] = []
    if subscribe_to_mountaineer:
        # Found mountaineer core and mountaineer external dependencies
        dependent_packages = find_packages_with_prefix("mountaineer")
        LOGGER.debug(
            f"Subscribing to changes in local mountaineer packages: {dependent_packages}"
        )

    return PackageWatchdog(
        client_package,
        dependent_packages=dependent_packages,
        callbacks=[
            CallbackDefinition(
                CallbackType.CREATED | CallbackType.MODIFIED,
                callback,
            )
        ],
        # We want to generate a build on the first load
        run_on_bootup=True,
    )


@contextmanager
def get_mountaineer_isolated_env(package: str):
    """
    Allow users to deny certain imports from being loaded into the core environment, in case
    these launch threads during initialization.

    :param package: The package to isolate imports from.

    """
    ignored_modules_raw = getenv("MOUNTAINEER_IGNORE_HOTRELOAD", "")
    ignored_modules = (
        [mod.strip() for mod in ignored_modules_raw.split(",")]
        if ignored_modules_raw
        else None
    )

    with isolate_imports(package, ignored_modules=ignored_modules) as environment:
        yield environment



from pathlib import Path

from mountaineer.controller import ControllerBase
from mountaineer.render import (
    RenderBase,
)


class StubRenderBase(RenderBase):
    pass


class StubController(ControllerBase):
    view_path = "/page.tsx"

    def render(self):
        return StubRenderBase()


def test_resolve_paths(tmp_path: Path):
    view_base = tmp_path / "views"
    ssr_base = view_base / "_ssr"
    static_base = view_base / "_static"

    controller = StubController()
    assert not controller.resolve_paths(view_base)

    # Now create an actual view path that we can sniff
    # This will get further in the pipeline but still won't be valid
    # because we don't have any of the global script files
    view_base.mkdir()
    assert not controller.resolve_paths(view_base)

    # Now we create the SSR script file
    ssr_base.mkdir()
    (ssr_base / "stub_controller.js").touch()
    (ssr_base / "stub_controller.js.map").touch()
    assert not controller.resolve_paths(view_base)

    # Our hash has to be exactly 32 digits to match the regex
    static_base.mkdir()
    random_hash = "b5ecd0c4405374100d6ef93088b86898"
    (static_base / f"stub_controller-{random_hash}.js").touch()
    (static_base / f"stub_controller-{random_hash}.js.map").touch()
    assert controller.resolve_paths(view_base)

    # Now ensure that the paths are correctly set
    assert controller._view_base_path == view_base
    assert controller._ssr_path == ssr_base / "stub_controller.js"
    assert controller._bundled_scripts == [f"stub_controller-{random_hash}.js"]



from click import command, option
from mountaineer.cli import handle_build, handle_runserver, handle_watch


@command()
@option("--host", default="127.0.0.1")
@option("--port", default=5006)
def runserver(host: str, port: int):
    handle_runserver(
        package="ci_webapp",
        webservice="ci_webapp.main:app",
        webcontroller="ci_webapp.app:controller",
        host=host,
        port=port,
        subscribe_to_mountaineer=True,
    )


@command()
def watch():
    handle_watch(
        package="ci_webapp",
        webcontroller="ci_webapp.app:controller",
        subscribe_to_mountaineer=True,
    )


@command()
def build():
    handle_build(
        webcontroller="ci_webapp.app:controller",
    )



from uuid import UUID, uuid4

from fastapi import Request
from mountaineer import ControllerBase, Metadata, RenderBase, passthrough, sideeffect
from pydantic import BaseModel


class HomeRender(RenderBase):
    client_ip: str
    current_count: int
    random_uuid: UUID


class IncrementCountRequest(BaseModel):
    count: int


class GetExternalDataResponse(BaseModel):
    first_name: str


class HomeController(ControllerBase):
    url = "/"
    view_path = "/app/home/page.tsx"

    def __init__(self):
        super().__init__()
        self.global_count = 0

    @sideeffect
    def increment_count(self, payload: IncrementCountRequest):
        self.global_count += payload.count

    @sideeffect(reload=(HomeRender.current_count,))
    def increment_count_only(self, payload: IncrementCountRequest, url_param: int):
        self.global_count += payload.count

    @passthrough(response_model=GetExternalDataResponse)
    def get_external_data(self) -> GetExternalDataResponse:
        # Execute a server action without automatically reloading the server state
        # Typically side-effects are the recommended way to get static data to the client
        # but sometimes you need to get data from a third-party API or gated resource
        return GetExternalDataResponse(
            first_name="John",
        )

    def render(self, request: Request) -> HomeRender:
        return HomeRender(
            client_ip=request.client.host if request.client else "unknown",
            current_count=self.global_count,
            # Bust the server-side cache
            random_uuid=uuid4(),
            metadata=Metadata(title="Home"),
        )



from time import sleep

import psycopg
from click import secho
from psycopg.errors import OperationalError

from create_mountaineer_app.generation import ProjectMetadata


def is_database_ready(metadata: ProjectMetadata):
    try:
        conn = psycopg.connect(
            # Default parameters specified in the templated .env
            dbname=f"{metadata.project_name}_db",
            user=f"{metadata.project_name}",
            password=metadata.postgres_password,
            host="localhost",
            port=metadata.postgres_port,
        )
        conn.close()
        return True
    except OperationalError:
        return False


def wait_for_database_to_be_ready(metadata: ProjectMetadata, max_wait=10):
    while max_wait > 0:
        if is_database_ready(metadata):
            secho("Database is ready!", fg="green")
            break

        max_wait -= 1
        secho("Waiting for database to become ready...", fg="yellow")
        sleep(1)

    if max_wait == 0:
        raise Exception("Timed out waiting for database to be ready.")



class CodeBlock:
    """
    Semantic grouping of a particular section of code, typically separated
    from other ones with two blank lines.

    """

    def __init__(self, lines: list[str]):
        self.lines = lines

    @staticmethod
    def indent(line: str):
        """
        Use the first line of the code block to determine the indentation level
        for the other lines that are implicitly imbedded in this line. This most
        acutely covers cases where string templating includes newlines so we can
        preserve the overall layout.

        ie. CodeBlock.indent(f"  my_var = {my_var}\n")

        """
        pass



from warnings import filterwarnings

import pytest

from mountaineer.config import unregister_config


@pytest.fixture(autouse=True)
def ignore_httpx_deprecation_warnings():
    # Ignore httpx deprecation warnings until fastapi updates its internal test constructor
    # https://github.com/encode/httpx/blame/master/httpx/_client.py#L678
    filterwarnings("ignore", category=DeprecationWarning, module="httpx.*")


@pytest.fixture(autouse=True)
def ignore_memory_object_resource_warnings():
    # @pierce - Fix a warning exposed in test_build_exception_on_get
    # from one of our underlying libraries
    # filterwarnings("ignore", category=ResourceWarning, module="anyio.*")
    filterwarnings("ignore", category=ResourceWarning)


@pytest.fixture(autouse=True)
def ignore_pluggy_warnings():
    filterwarnings(
        "ignore",
        category=pytest.PytestWarning,
        message=".*PluggyTeardownRaisedWarning.*",
    )


@pytest.fixture(autouse=True)
def clear_config_cache():
    unregister_config()



{% if create_stub_files %}
from .home import HomeController as HomeController
from .detail import DetailController as DetailController
{% else %}
# Specify sub-controllers here
# Ex: from .home import HomeController
{% endif %}


import importlib
from pathlib import Path
from tempfile import mkdtemp
from traceback import format_exception
from typing import TYPE_CHECKING

from fastapi import Request

from mountaineer.app import AppController
from mountaineer.client_builder.builder import APIBuilder
from mountaineer.client_compiler.compile import ClientCompiler
from mountaineer.development.messages import (
    BootupMessage,
    BuildJsMessage,
    BuildUseServerMessage,
    ErrorResponse,
    MessageTypes,
    StartServerMessage,
    SuccessResponse,
)
from mountaineer.development.messages_broker import AsyncMessageBroker
from mountaineer.development.uvicorn import UvicornThread
from mountaineer.logging import setup_internal_logger

if TYPE_CHECKING:
    from mountaineer_exceptions.controllers.exception_controller import (
        ExceptionController,
    )

LOGGER = setup_internal_logger(__name__)


class IsolatedAppContext:
    """
    Manages the app controller, server, and compilation.

    """

    def __init__(
        self,
        package: str,
        package_path: Path,
        module_name: str,
        controller_name: str,
        use_dev_exceptions: bool = True,
    ):
        """
        Initialize an isolated application context in a separate process.

        :param package: Name of the main package
        :param package_path: Path to the package on disk
        :param module_name: Module name containing the app controller
        :param controller_name: Variable name of the app controller within the module
        """
        super().__init__()
        self.package = package
        self.package_path = package_path
        self.module_name = module_name
        self.controller_name = controller_name
        self.webservice_thread: UvicornThread | None = None

        self.app_controller: AppController | None = None
        self.exception_controller: "ExceptionController | None" = None
        self.use_dev_exceptions = use_dev_exceptions

        self.js_compiler: APIBuilder | None = None
        self.app_compiler: ClientCompiler | None = None

    @classmethod
    def from_webcontroller(cls, webcontroller: str, use_dev_exceptions: bool = True):
        package = webcontroller.split(".")[0]
        module_name = webcontroller.split(":")[0]
        controller_name = webcontroller.split(":")[1]

        return cls(
            package=package,
            package_path=Path(package.replace(".", "/")),
            module_name=module_name,
            controller_name=controller_name,
            use_dev_exceptions=use_dev_exceptions,
        )

    async def run_async(self, broker: AsyncMessageBroker[MessageTypes]):
        """
        Asynchronous main loop that processes messages from the broker.

        This is the core message handling loop that:
        1. Retrieves messages from the queue
        2. Dispatches them to appropriate handlers
        3. Sends responses back through the broker

        The loop continues until a ShutdownMessage is received.
        """
        try:
            LOGGER.debug("[IsolatedAppContext] Starting isolated context")

            # Process messages until shutdown
            while True:
                LOGGER.debug("Will block on next message")
                message_id, message = await broker.get_job()
                LOGGER.debug(f"Got message: {message}")

                try:
                    response: SuccessResponse | ErrorResponse
                    if isinstance(message, BootupMessage):
                        response = await self.initialize_app_state()
                    elif isinstance(message, StartServerMessage):
                        response = await self.handle_start_server(
                            host=message.host,
                            port=message.port,
                            live_reload_port=message.live_reload_port,
                        )
                    elif isinstance(message, BuildJsMessage):
                        response = await self.handle_js_build(message.updated_js)
                    elif isinstance(message, BuildUseServerMessage):
                        response = await self.handle_build_use_server()
                    else:
                        LOGGER.error(f"Invalid message type: {type(message)} {message}")
                        raise ValueError(
                            f"Invalid message type: {type(message)} {message}"
                        )
                    LOGGER.debug(f"Will write response: {response}")
                    await broker.send_response(message_id, response)
                except Exception as e:
                    LOGGER.debug(
                        f"Isolated app context failed to process message: {e}, continuing...",
                        exc_info=True,
                    )
                    await broker.send_response(
                        message_id,
                        ErrorResponse(
                            exception=str(e), traceback="".join(format_exception(e))
                        ),
                    )
        except Exception as e:
            LOGGER.error(f"Isolated app context failed: {e}", exc_info=True)
        finally:
            if self.webservice_thread:
                await self.webservice_thread.astop()

    #
    # Message Handlers
    #

    async def handle_build_use_server(self):
        """
        Build the useServer support files for client-side React Server Components.

        :return: Success response on successful build
        :raises ValueError: If JS compiler is not initialized
        """
        if self.js_compiler is None:
            raise ValueError("JS compiler not initialized")

        await self.js_compiler.build_use_server()
        return SuccessResponse()

    async def handle_js_build(self, updated_js: list[Path] | None = None):
        """
        Compile JavaScript files based on updated paths.

        Runs the builder plugins and invalidates any affected views in the app controller.

        :param updated_js: Optional list of specific JS files that were updated
        :return: Success response on successful build
        :raises ValueError: If app compiler or controller is not initialized
        """
        if self.app_compiler is None:
            raise ValueError("App compiler not initialized")
        if self.app_controller is None:
            raise ValueError("App controller not initialized")

        await self.app_compiler.run_builder_plugins(limit_paths=updated_js)
        for path in updated_js or []:
            self.app_controller.invalidate_view(path)

        return SuccessResponse()

    #
    # Server Initialization
    #

    async def initialize_app_state(self):
        """
        Initialize all application state components within the isolated context.

        This method:
        1. Loads the web service and app controller
        2. Initializes the hot reloader
        3. Mounts the exception controller for development error pages
        4. Sets up JS and app compilers

        :return: Success response on successful initialization
        :raises ValueError: If app controller fails to initialize
        """
        # Import and initialize the module
        self.module = importlib.import_module(self.module_name)
        initial_state = {name: getattr(self.module, name) for name in dir(self.module)}
        self.app_controller = initial_state[self.controller_name]

        if self.app_controller is None:
            raise ValueError("App controller not initialized")

        # Mount exceptions
        self.mount_exceptions(self.app_controller)

        # Initialize builders in isolated context
        global_build_cache = Path(mkdtemp())
        self.js_compiler = APIBuilder(
            self.app_controller,
            build_cache=global_build_cache,
        )
        self.app_compiler = ClientCompiler(
            app=self.app_controller,
        )

        return SuccessResponse()

    async def handle_start_server(self, host: str, port: int, live_reload_port: int):
        """
        Start the Uvicorn server for the web application.

        Configures and launches a UvicornThread to serve the application.
        If a server is already running, it will be stopped first.

        :raises ValueError: If app controller is not initialized
        """
        if self.webservice_thread is not None:
            LOGGER.debug("Server is already running")
            return SuccessResponse()

        if self.app_controller is None:
            raise ValueError("App controller not initialized")

        # Inject the live reload port
        self.app_controller.live_reload_port = live_reload_port or 0

        self.webservice_thread = UvicornThread(
            name="Dev webserver",
            emoticon="🚀",
            app=self.app_controller.app,
            host=host,
            port=port,
        )
        await self.webservice_thread.astart()

        return SuccessResponse()

    #
    # Dev Hooks
    #

    def mount_exceptions(self, app_controller: AppController):
        """
        Mount the exception controller to the app controller for custom error pages.

        This adds development-friendly error pages with stack traces and debugging information.
        It ensures the exception controller is only mounted once.

        :param app_controller: The app controller to mount the exception controller on
        """
        if not self.use_dev_exceptions:
            LOGGER.debug("Dev exceptions are disabled, skipping...")
            return

        try:
            from mountaineer_exceptions.controllers.exception_controller import (
                ExceptionController,
            )
            from mountaineer_exceptions.plugin import plugin  # type: ignore

            app_controller.register(plugin)
            self.exception_controller = [
                controller
                for controller in plugin.get_controllers()
                if isinstance(controller, ExceptionController)
            ][0]
            app_controller.app.exception_handler(Exception)(self.handle_dev_exception)
        except ImportError:
            LOGGER.warning("mountaineer-exceptions plugin not found, skipping...")

    async def handle_dev_exception(self, request: Request, exc: Exception):
        """
        Handle exceptions in development mode with enhanced error pages.

        For GET requests, renders a detailed error page with stack traces and debugging info.
        For other request types, re-raises the exception to be handled by the framework.

        :param request: The FastAPI request object
        :param exc: The exception that was raised
        :return: HTML response for GET requests
        :raises: The original exception for non-GET requests
        """
        LOGGER.error(f"Handling dev exception: {exc}")

        if not self.exception_controller:
            raise ValueError("Exception controller not initialized")

        if request.method == "GET":
            html = await self.exception_controller._definition.view_route(  # type: ignore
                exception=str(exc),
                stack="".join(format_exception(exc)),
                parsed_exception=self.exception_controller.traceback_parser.parse_exception(
                    exc
                ),
            )
            return html
        else:
            raise exc



from pathlib import Path

import pytest

from mountaineer.app import AppController
from mountaineer.controller_layout import LayoutControllerBase


class ExampleLayoutController(LayoutControllerBase):
    view_path = "/test.tsx"

    async def render(self) -> None:
        pass


@pytest.mark.asyncio
async def test_disallows_direct_rendering():
    """
    Layouts can't be rendered directly. They're intended for use
    only as pages wrappers.

    """
    layout_controller = ExampleLayoutController()

    with pytest.raises(NotImplementedError):
        await layout_controller._generate_html(global_metadata={})

    with pytest.raises(NotImplementedError):
        layout_controller._generate_ssr_html({})


def test_layout_registration():
    """
    Layout controllers must register with the AppController for them
    to be used by client render controllers.

    """
    app_controller = AppController(view_root=Path(""))
    layout_controller = ExampleLayoutController()

    app_controller.register(layout_controller)
    assert len(app_controller.controllers) == 1

    assert app_controller.controllers[0].controller == layout_controller



from inspect import getsource
from pathlib import Path
from textwrap import dedent
from typing import Any, AsyncIterator, Iterator, cast

import mypy.api
import pytest
from fastapi.responses import JSONResponse, StreamingResponse
from fastapi.testclient import TestClient
from pydantic import BaseModel

from mountaineer.actions.fields import FunctionActionType, get_function_metadata
from mountaineer.actions.passthrough_dec import (
    passthrough,
)
from mountaineer.annotation_helpers import MountaineerUnsetValue
from mountaineer.app import AppController
from mountaineer.controller import ControllerBase
from mountaineer.logging import LOGGER
from mountaineer.render import RenderBase


def test_markup_passthrough():
    """
    Check that the @passthrough decorator extracts the expected
    data from our model definition.
    """

    class ExamplePassthroughModel(BaseModel):
        first_name: str

    class ExampleController(ControllerBase):
        view_path = "/test.tsx"

        @passthrough
        def get_external_data(self) -> ExamplePassthroughModel:
            return ExamplePassthroughModel(
                first_name="John",
            )

    metadata = get_function_metadata(ExampleController.get_external_data)
    assert metadata.action_type == FunctionActionType.PASSTHROUGH
    assert metadata.get_passthrough_model() == ExamplePassthroughModel
    assert metadata.function_name == "get_external_data"
    assert isinstance(metadata.reload_states, MountaineerUnsetValue)
    assert isinstance(metadata.render_model, MountaineerUnsetValue)


class ExampleRenderModel(RenderBase):
    value_a: str
    value_b: str


class ExamplePassthroughModel(BaseModel):
    status: str


class ExampleController(ControllerBase):
    url: str = "/test/{query_id}/"
    view_path = "/test.tsx"

    def __init__(self):
        super().__init__()
        self.counter = 0
        self.render_counts = 0

    def render(
        self,
        query_id: int,
    ) -> ExampleRenderModel:
        self.render_counts += 1
        return ExampleRenderModel(
            value_a="Hello",
            value_b="World",
        )

    @passthrough
    def call_passthrough(self, payload: dict) -> ExamplePassthroughModel:
        self.counter += 1
        return ExamplePassthroughModel(status="success")

    @passthrough
    async def call_passthrough_async(self, payload: dict) -> ExamplePassthroughModel:
        self.counter += 1
        return ExamplePassthroughModel(status="success")


@pytest.mark.asyncio
async def test_can_call_passthrough():
    app = AppController(view_root=Path())
    controller = ExampleController()
    app.register(controller)

    return_value_sync = await controller.call_passthrough(
        {},
    )

    return_value_async = await controller.call_passthrough_async(
        {},
    )

    # The response payload should be the same both both sync and async endpoints
    expected_response = {
        "passthrough": ExamplePassthroughModel(
            status="success",
        )
    }

    assert return_value_sync == expected_response
    assert return_value_async == expected_response

    assert controller.counter == 2

    # Our passthrough logic by definition should not re-render
    assert controller.render_counts == 0


@pytest.mark.asyncio
async def test_can_call_passthrough_original():
    """
    Ensure that we can access the raw underlying function that was
    wrapped by the decorator.

    """
    controller = ExampleController()
    assert await ExampleController.call_passthrough.original(
        controller, dict()
    ) == ExamplePassthroughModel(status="success")
    assert await ExampleController.call_passthrough_async.original(
        controller, dict()
    ) == ExamplePassthroughModel(status="success")


class ExampleModel(BaseModel):
    value: str


class ExampleIterableController(ControllerBase):
    url = "/example"
    view_path = "/test.tsx"

    async def render(self) -> None:
        pass

    @passthrough
    async def get_data(self) -> AsyncIterator[ExampleModel]:
        yield ExampleModel(value="Hello")
        yield ExampleModel(value="World")


def test_extracts_iterable():
    controller = ExampleIterableController()
    metadata = get_function_metadata(controller.get_data)
    assert metadata.passthrough_model == ExampleModel
    # Explicitly validate type here instead of using global constant
    assert metadata.media_type == "text/event-stream"


def test_disallows_invalid_iterables():
    # Sync functions
    with pytest.raises(ValueError, match="async generators are supported"):

        class ExampleController1(ControllerBase):
            @passthrough  # type: ignore
            def sync_iterable(self) -> Iterator[ExampleModel]:
                yield ExampleModel(value="Hello")
                yield ExampleModel(value="World")

    # Generator without marking up the response model
    with pytest.raises(ValueError, match="must have a response_model"):

        class ExampleController2(ControllerBase):
            @passthrough
            async def no_response_type_iterable(self) -> None:  # type: ignore
                yield ExampleModel(value="Hello")  # type: ignore
                yield ExampleModel(value="World")  # type: ignore


@pytest.mark.asyncio
async def test_can_call_iterable():
    app = AppController(view_root=Path())
    controller = ExampleIterableController()
    app.register(controller)

    # Ensure we return a valid StreamingResponse when called directly from the code
    return_value_sync = cast(Any, await controller.get_data())
    assert isinstance(return_value_sync, StreamingResponse)

    # StreamingResponses are intended to be read by an ASGI server, so we'll use the TestClient to simulate one instead of calling directly
    controller_definition = app._definition_for_controller(controller)
    passthrough_url = controller_definition.get_url_for_metadata(
        get_function_metadata(controller.get_data)
    )

    client = TestClient(app.app)
    lines: list[str] = []
    with client.stream(
        "POST",
        passthrough_url,
        json={},
    ) as response:
        for line in response.iter_lines():
            lines.append(line)

    assert lines == [
        'data: {"passthrough": {"value": "Hello"}}',
        'data: {"passthrough": {"value": "World"}}',
    ]


@pytest.mark.asyncio
async def test_raw_response():
    class ExampleController(ControllerBase):
        url: str = "/test/{query_id}/"
        view_path = "/test.tsx"

        def render(
            self,
            query_id: int,
        ) -> ExampleRenderModel:
            return ExampleRenderModel(value_a="Hello", value_b="World")

        @passthrough(raw_response=True)
        def call_passthrough(self, payload: dict) -> JSONResponse:
            return JSONResponse(content={"raw_value": "success"})

    app = AppController(view_root=Path())
    controller = ExampleController()
    app.register(controller)

    controller_definition = app._definition_for_controller(controller)

    client = TestClient(app.app)
    response = client.post(
        controller_definition.get_url_for_metadata(
            get_function_metadata(controller.call_passthrough)
        ),
        json={},
    )
    # No "passthrough" wrapping
    assert response.json() == {"raw_value": "success"}


@pytest.mark.parametrize(
    "passthrough_value, return_typehint, return_value, is_valid",
    [
        # Normal return type
        ("@passthrough", "InputModel", "return InputModel()", True),
        (
            "@passthrough(exception_models=[])",
            "InputModel",
            "return InputModel()",
            True,
        ),
        # Raw responses have to be fastapi.Response models or subclasses
        ("@passthrough(raw_response=True)", "InputModel", "return InputModel()", False),
        (
            "@passthrough(raw_response=True)",
            "JSONResponse",
            "return JSONResponse(content={})",
            True,
        ),
        (
            "@passthrough(raw_response=True)",
            "HTMLResponse",
            'return HTMLResponse(content="TEST")',
            True,
        ),
        # We can return JSONResponse is normal passthrough functions as well, because they'll be wrapped
        # in a passthrough response
        # We can't return other fastapi.Response models, however
        ("@passthrough", "JSONResponse", "return JSONResponse(content={})", True),
        ("@passthrough", "HTMLResponse", 'return HTMLResponse(content="TEST")', False),
        # Iterator types are only allowed if they're async
        ("@passthrough", "AsyncIterator[InputModel]", "yield InputModel()", True),
    ],
)
def test_passthrough_typechecking(
    passthrough_value: str,
    return_typehint: str,
    return_value: str,
    is_valid: bool,
    tmp_path: Path,
):
    """
    Ensure that mypy will catch type errors in passthrough signatures

    """

    def run_function():
        from typing import AsyncIterator  # noqa: F401

        from fastapi.responses import HTMLResponse, JSONResponse  # noqa: F401
        from pydantic import BaseModel  # noqa: F401

        from mountaineer import ControllerBase, passthrough  # noqa: F401

        class InputModel(BaseModel):
            pass

        class TestController:
            @passthrough  # type: ignore
            async def get_external_data(self) -> str:
                return "{output_value}"

    # Ignore the "run_function()" function header itself
    function_lines = getsource(run_function).split("\n")[1:]
    value = dedent("\n".join(function_lines))

    value = value.replace("@passthrough", passthrough_value)
    value = value.replace("-> str", f"-> {return_typehint}")
    value = value.replace('return "{output_value}"', return_value)
    value = value.replace("# type: ignore", "")

    LOGGER.debug(f"Input value:\n{value}")

    module_path = tmp_path / "test.py"
    module_path.write_text(value)

    result = mypy.api.run([str(module_path)])
    LOGGER.debug(f"mypy result: {result}")

    if is_valid:
        assert "Success" in result[0]
    else:
        assert "error" in result[0]



from typing import Any, TypeVar, get_args, get_origin


def get_typevar_mapping(cls):
    """
    Get the raw typevar mappings {typvar: generic values} for each
    typevar in the class hierarchy of `cls`.

    """
    mapping: dict[Any, Any] = {}

    # Traverse MRO in reverse order, except `object`
    for base in reversed(cls.__mro__[:-1]):
        # Skip non-generic classes
        if not hasattr(base, "__orig_bases__"):
            continue

        for origin_base in base.__orig_bases__:
            origin = get_origin(origin_base)
            if origin:
                base_params = getattr(origin, "__parameters__", [])
                instantiated_params = get_args(origin_base)

                # Update mapping with current base's mappings
                base_mapping = dict(zip(base_params, instantiated_params))
                for key, value in base_mapping.items():
                    # If value is another TypeVar, resolve it if possible
                    if isinstance(value, TypeVar) and value in mapping:
                        mapping[key] = mapping[value]
                    else:
                        mapping[key] = value

    # Exclude TypeVars from the final mapping
    return mapping


def expand_typevars(raw_values):
    """
    Expand all typevars in the class hierarchy of `cls` to their
    final values. This resolves cases where a typevar in an ancestor
    is resolved by a child value.

    """
    final_values: dict[Any, Any] = {}

    for key, value in raw_values.items():
        # Resolve until we can't resolve anymore
        while isinstance(value, TypeVar) and value in raw_values:
            value = raw_values[value]

        final_values[key] = value

    return final_values



from contextlib import contextmanager
from os import chdir
from pathlib import Path

import pytest
import toml

from scripts.update_version import (
    format_cargo_version,
    format_python_version,
    update_version_python,
    update_version_rust,
)


@contextmanager
def change_pwd(new_pwd: Path):
    old_pwd = Path.cwd()
    try:
        chdir(new_pwd)
        yield
    finally:
        chdir(old_pwd)


@pytest.mark.parametrize(
    "input_version,expected",
    [
        ("1.2.3", "1.2.3"),
        ("1.2", "1.2.0"),
        ("1.2.3-alpha.1", "1.2.3-a1"),
        ("1.2.3-beta.2", "1.2.3-b2"),
        ("1.2.3.post1", "1.2.3-post1"),
        ("1.2.3.dev4", "1.2.3-dev4"),
        ("1.2.3-alpha.1.post2.dev3", "1.2.3-a1-post2-dev3"),
    ],
)
def test_format_cargo_version(input_version, expected):
    assert format_cargo_version(input_version) == expected


def test_update_version_rust(tmp_path: Path):
    cargo_text = """
        [package]
        name = "mountaineer"
        # Bumped automatically by CI on a release
        version = "0.1.0"
        edition = "2021"

        [dependencies]
        v8 = "0.89.0"
        deno_core_icudata = "0.73.0"
        """

    cargo_path = tmp_path / "Cargo.toml"
    cargo_path.write_text(cargo_text)

    with change_pwd(tmp_path):
        update_version_rust("0.2.0")

    assert toml.loads(cargo_path.read_text()) == {
        "package": {
            "name": "mountaineer",
            "version": "0.2.0",
            "edition": "2021",
        },
        "dependencies": {
            "v8": "0.89.0",
            "deno_core_icudata": "0.73.0",
        },
    }


@pytest.mark.parametrize(
    "input_version,expected",
    [
        ("1.2.3", "1.2.3"),
        ("1.2.3a1", "1.2.3b1"),
        ("1.2.3b2", "1.2.3b2"),
        ("1.2.3rc3", "1.2.3b3"),
        ("1.2.3.post1", "1.2.3.post1"),
        ("1.2.3.dev4", "1.2.3.dev4"),
        ("1.2.3a1.post2.dev3", "1.2.3b1.post2.dev3"),
    ],
)
def test_format_python_version(input_version, expected):
    assert format_python_version(input_version) == expected


def test_update_version_python(tmp_path: Path):
    pyproject_text = """
        [tool.poetry]
        name = "mountaineer"
        version = "0.1.0"
        description = ""
        readme = "README.md"

        [tool.poetry.dependencies]
        uvicorn = { extras = ["standard"], version = "^0.27.0.post1" }
        fastapi = "^0.68.0"
        """

    pyproject_path = tmp_path / "pyproject.toml"
    pyproject_path.write_text(pyproject_text)

    with change_pwd(tmp_path):
        update_version_python("0.2.0")

    assert toml.loads(pyproject_path.read_text()) == {
        "tool": {
            "poetry": {
                "name": "mountaineer",
                "version": "0.2.0",
                "description": "",
                "readme": "README.md",
                "dependencies": {
                    "uvicorn": {
                        "extras": ["standard"],
                        "version": "^0.27.0.post1",
                    },
                    "fastapi": "^0.68.0",
                },
            },
        },
    }


def test_update_version_python_with_project_version(tmp_path: Path):
    pyproject_text = """
        [tool.poetry]
        name = "mountaineer"
        version = "0.1.0"
        description = ""
        readme = "README.md"

        [project]
        name = "mountaineer"
        version = "0.1.0"
        description = ""

        [tool.poetry.dependencies]
        uvicorn = { extras = ["standard"], version = "^0.27.0.post1" }
        fastapi = "^0.68.0"
        """

    pyproject_path = tmp_path / "pyproject.toml"
    pyproject_path.write_text(pyproject_text)

    with change_pwd(tmp_path):
        update_version_python("0.2.0")

    assert toml.loads(pyproject_path.read_text()) == {
        "tool": {
            "poetry": {
                "name": "mountaineer",
                "version": "0.2.0",
                "description": "",
                "readme": "README.md",
                "dependencies": {
                    "uvicorn": {
                        "extras": ["standard"],
                        "version": "^0.27.0.post1",
                    },
                    "fastapi": "^0.68.0",
                },
            },
        },
        "project": {
            "name": "mountaineer",
            "version": "0.2.0",
            "description": "",
        },
    }


def test_update_version_python_only_project(tmp_path: Path):
    pyproject_text = """
        [project]
        name = "mountaineer"
        version = "0.1.0"
        description = ""
        """

    pyproject_path = tmp_path / "pyproject.toml"
    pyproject_path.write_text(pyproject_text)

    with change_pwd(tmp_path):
        update_version_python("0.2.0")

    assert toml.loads(pyproject_path.read_text()) == {
        "project": {
            "name": "mountaineer",
            "version": "0.2.0",
            "description": "",
        },
    }


def test_update_version_python_only_poetry(tmp_path: Path):
    pyproject_text = """
        [tool.poetry]
        name = "mountaineer"
        version = "0.1.0"
        description = ""
        """

    pyproject_path = tmp_path / "pyproject.toml"
    pyproject_path.write_text(pyproject_text)

    with change_pwd(tmp_path):
        update_version_python("0.2.0")

    assert toml.loads(pyproject_path.read_text()) == {
        "tool": {
            "poetry": {
                "name": "mountaineer",
                "version": "0.2.0",
                "description": "",
            },
        },
    }


def test_update_version_python_no_version_sections(tmp_path: Path, capsys):
    pyproject_text = """
        [build-system]
        requires = ["poetry-core"]
        build-backend = "poetry.core.masonry.build"
        """

    pyproject_path = tmp_path / "pyproject.toml"
    pyproject_path.write_text(pyproject_text)

    with change_pwd(tmp_path):
        update_version_python("0.2.0")

    # Verify the warning was printed
    captured = capsys.readouterr()
    assert "Neither [tool.poetry] nor [project] sections found" in captured.out

    # Verify the file wasn't modified
    assert toml.loads(pyproject_path.read_text()) == {
        "build-system": {
            "requires": ["poetry-core"],
            "build-backend": "poetry.core.masonry.build",
        },
    }



from contextlib import contextmanager

from pydantic._internal._model_construction import ModelMetaclass
from pydantic.fields import Field
from pydantic_settings import BaseSettings
from typing_extensions import dataclass_transform


class ConfigMeta(ModelMetaclass):
    def __call__(cls, *args, **kwargs):
        instance = super().__call__(*args, **kwargs)
        register_config(instance)
        return instance


@dataclass_transform(kw_only_default=True, field_specifiers=(Field,))
class ConfigBase(BaseSettings, metaclass=ConfigMeta):
    """
    Base class for the running application's configuration. By convention
    all configuration parameters should be specified here in one payload.
    You'll often call your subclass `ConfigBase`.

    Users are responsible for instantiating an AppConfig with your desired
    settings. This instance will be registered into the global space so it's
    accessible to your controllers. An error will be thrown if you attempt to
    instantiate more than one AppConfig.

    """

    # Name of the python package. Will be used to sniff for the installed
    # codebase in the current virtualenv.
    PACKAGE: str | None = None

    # Environment flag. Set to anything you want. Only if set to "development" will
    # we include frontend artifacts for the hot-reloading server.
    ENVIRONMENT: str = "development"

    model_config = {"frozen": True}


# One global config object
APP_CONFIG: ConfigBase | None = None


def register_config(config: ConfigBase):
    """
    Manually register a configuration instance into the global space. Each application
    can have a maximum of one configuration instance registered. If you attempt to
    register a second instance, an error will be thrown.

    Registration should happen automatically by initializing a new instance of your
    ConfigBase class on application start. This auto-registration is provided by
    the configuration's metaclass in ConfigMeta.

    """
    global APP_CONFIG

    if APP_CONFIG is not None and APP_CONFIG != config:
        raise ValueError("Config already registered")

    APP_CONFIG = config


def unregister_config():
    """
    Unregister the current configuration instance.

    """
    global APP_CONFIG
    APP_CONFIG = None


def get_config() -> ConfigBase:
    """
    Get the current configuration instance that's registered globally. Will
    throw an error if no configuration instance is registered.

    """
    if APP_CONFIG is None:
        raise ValueError(
            "Configuration not registered. Either:\n"
            "1. Call register_config() with your BaseSettings class\n"
            "2. Make sure your BaseSettings is imported so the ConfigMeta can auto-register"
        )

    return APP_CONFIG


@contextmanager
def register_config_in_context(config: ConfigBase):
    """
    Change the global config object to the given config object
    temporarily. Useful for unit testing.

    """
    global APP_CONFIG
    previous_config = APP_CONFIG
    APP_CONFIG = config
    try:
        yield
    finally:
        APP_CONFIG = previous_config



import asyncio
import json
import pickle
import secrets
import socket
from asyncio import Future
from base64 import b64decode, b64encode
from contextlib import asynccontextmanager
from dataclasses import dataclass
from threading import Thread
from typing import Annotated, Any, Generic, Literal, TypeVar, cast
from uuid import uuid4

from pydantic import BaseModel, Field, TypeAdapter

from mountaineer.development.messages import ErrorResponse
from mountaineer.logging import LOGGER

TResponse = TypeVar("TResponse")
AppMessageTypes = TypeVar("AppMessageTypes")


class BrokerMessageFuture(Generic[TResponse], asyncio.Future[TResponse]):
    pass


@dataclass
class BrokerServerConfig(Generic[AppMessageTypes]):
    """
    Config required to access a server that controls queues
    in a separate thread.
    """

    host: str
    port: int
    auth_key: str


class BaseCommand(BaseModel):
    """Base class for all broker commands with authentication"""

    auth_key: str


class SendJobCommand(BaseCommand):
    """Command to send a new job to the broker"""

    item_type: Literal["send_job"] = "send_job"
    job_id: str
    job_data: Any


class SendResponseCommand(BaseCommand):
    """Command to send a response for a job"""

    item_type: Literal["send_response"] = "send_response"
    job_id: str
    response_data: Any


class GetResponseCommand(BaseCommand):
    """Command to get a response for a job"""

    item_type: Literal["get_response"] = "get_response"
    job_id: str


class GetJobCommand(BaseCommand):
    """Command to get the next available job"""

    item_type: Literal["get_job"] = "get_job"


class DrainQueueCommand(BaseCommand):
    """Command to drain all jobs from the queue"""

    item_type: Literal["drain_queue"] = "drain_queue"


class BaseResponse(BaseModel):
    """Base class for all broker responses"""


class OKResponse(BaseResponse):
    """Response indicating successful command execution"""

    item_type: Literal["ok"] = "ok"
    response_data: Any | None = None


class UnauthorizedResponse(BaseResponse):
    """Response indicating authentication failure"""

    item_type: Literal["error"] = "error"
    message: str = "Invalid authentication key"


class BrokerAuthenticationError(Exception):
    """Raised when the broker authentication fails."""

    pass


class BrokerExecutionError(Exception):
    """Raised when the broker execution fails."""

    def __init__(self, error: str, traceback: str):
        super().__init__(error)
        self.error = error
        self.traceback = traceback


CommandTypes = (
    SendJobCommand
    | SendResponseCommand
    | GetResponseCommand
    | GetJobCommand
    | DrainQueueCommand
)
ResponseTypes = OKResponse | UnauthorizedResponse

# Create type adapters for polymorphic validation
command_type_adapter = TypeAdapter(  # type: ignore
    Annotated[CommandTypes, Field(discriminator="item_type")]
)

response_type_adapter = TypeAdapter(  # type: ignore
    Annotated[ResponseTypes, Field(discriminator="item_type")]
)


class AsyncMessageBroker(Thread, Generic[AppMessageTypes]):
    """
    A simple process-independent message broker server. This works around limitations
    with `multiprocessing.Queue` that require all processes to be related to the central
    running parent process (for sharing of file descriptors). Since exec processes are
    launched separately by firehot this inheritance isn't possible.

    On Linux we could work around this by using `mkfifo` but since there's no Windows
    equivalent for it, we use a simple port-based server to keep things multi-platform.

    """

    def __init__(
        self, *, host: str = "127.0.0.1", port: int = 0, auth_key: str | None = None
    ):
        """
        :param port: If set to 0, the server will choose a free port.
        :param auth_key: If provided, the server will check that the client's auth key matches.
        """
        super().__init__()
        self.host = host
        self.port = port
        self.auth_key = auth_key or secrets.token_hex(16)
        self.jobs: dict[str, Any] = {}  # job_id -> job_data
        self.job_queue: list[str] = []  # FIFO queue of job_ids
        self.responses: dict[str, Any] = {}  # job_id -> response_data
        self.pending_futures: dict[
            str, list[Future]
        ] = {}  # job_id -> list of asyncio.Future waiting for a response
        self.pending_job_futures: list[
            Future
        ] = []  # list of futures waiting for next job

        self.loop: asyncio.AbstractEventLoop | None = None
        self.server: asyncio.AbstractServer | None = None

        # Thread control.
        self.is_running = False
        self.should_stop = False

    def run(self):
        # Create and set the event loop for this thread.
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        # Determine a free port if needed.
        if self.port == 0:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind((self.host, 0))
                self.port = s.getsockname()[1]

        # Start the asyncio TCP server.
        coro = asyncio.start_server(self.handle_client, self.host, self.port)
        self.server = self.loop.run_until_complete(coro)
        LOGGER.info(f"JobBrokerServer started on {self.host}:{self.port}")

        self.is_running = True

        try:
            self.loop.run_forever()
        finally:
            # Clean up
            if self.server:
                self.server.close()
                self.loop.run_until_complete(self.server.wait_closed())

            # Cancel all pending tasks
            pending = asyncio.all_tasks(self.loop)
            for task in pending:
                task.cancel()

            # Wait for all tasks to complete with a timeout
            if pending:
                self.loop.run_until_complete(asyncio.wait(pending, timeout=5.0))

            self.loop.close()
            self.is_running = False

    async def stop_server(self):
        """Stop the broker server and clean up resources."""
        if not self.is_running:
            return
        if not self.loop:
            raise RuntimeError("Broker server async loop not running")

        self.should_stop = True

        # Cancel all pending futures
        for futures in self.pending_futures.values():
            for future in futures:
                if not future.done():
                    future.cancel()
        self.pending_futures.clear()

        # Cancel pending job futures
        for future in self.pending_job_futures:
            if not future.done():
                future.cancel()
        self.pending_job_futures.clear()

        if self.server:
            try:
                self.server.close()
            except TypeError:
                # Expected error during failed _wakeup in base_events:close()
                pass
            # Schedule wait_closed on the broker's loop
            fut = asyncio.run_coroutine_threadsafe(self.server.wait_closed(), self.loop)
            await asyncio.wrap_future(fut)

        if self.loop and self.loop.is_running():
            self.loop.call_soon_threadsafe(self.loop.stop)

    async def handle_client(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        """
        Handle incoming client connections using a JSON-over-TCP protocol.
        Each JSON message corresponds to one of our command dataclasses.
        """
        try:
            while not self.should_stop:
                try:
                    data = await reader.readline()
                    if not data:
                        break  # connection closed
                except (ConnectionError, asyncio.CancelledError):
                    break

                response: BaseResponse | None = None

                if not self.loop:
                    raise RuntimeError("Broker server async loop not running")

                try:
                    # Deserialize incoming JSON message using the type adapter
                    message_dict = json.loads(data.decode())
                    try:
                        cmd_obj = command_type_adapter.validate_python(message_dict)
                    except Exception as e:
                        raise ValueError(f"Invalid command format: {e}")

                    # Validate auth key
                    if cmd_obj.auth_key != self.auth_key:
                        response = UnauthorizedResponse()
                        writer.write(
                            (json.dumps(response.model_dump()) + "\n").encode()
                        )
                        await writer.drain()
                        continue

                    # Process command based on its type
                    if isinstance(cmd_obj, SendJobCommand):
                        self.jobs[cmd_obj.job_id] = cmd_obj.job_data
                        self.job_queue.append(cmd_obj.job_id)
                        # If any futures are waiting for jobs, notify one of them
                        if self.pending_job_futures:
                            fut = self.pending_job_futures.pop(0)
                            if not fut.done():
                                fut.set_result(
                                    {
                                        "job_id": cmd_obj.job_id,
                                        "job_data": cmd_obj.job_data,
                                    }
                                )
                        response = OKResponse()
                    elif isinstance(cmd_obj, SendResponseCommand):
                        self.responses[cmd_obj.job_id] = cmd_obj.response_data
                        # If any futures are waiting, notify them
                        if cmd_obj.job_id in self.pending_futures:
                            for fut in self.pending_futures[cmd_obj.job_id]:
                                if not fut.done():
                                    fut.set_result(cmd_obj.response_data)
                            del self.pending_futures[cmd_obj.job_id]
                        response = OKResponse()
                    elif isinstance(cmd_obj, GetResponseCommand):
                        if cmd_obj.job_id in self.responses:
                            response = OKResponse(
                                response_data=self.responses[cmd_obj.job_id]
                            )
                        else:
                            fut = self.loop.create_future()
                            self.pending_futures.setdefault(cmd_obj.job_id, []).append(
                                fut
                            )
                            response_data = await fut
                            response = OKResponse(response_data=response_data)
                    elif isinstance(cmd_obj, GetJobCommand):
                        if self.job_queue:
                            # Get next job from queue
                            job_id = self.job_queue.pop(0)
                            response = OKResponse(
                                response_data={
                                    "job_id": job_id,
                                    "job_data": self.jobs[job_id],
                                }
                            )
                        else:
                            # No jobs available, create a future to wait for one
                            fut = self.loop.create_future()
                            self.pending_job_futures.append(fut)
                            try:
                                job_info = await fut
                                response = OKResponse(response_data=job_info)
                            except asyncio.CancelledError:
                                response = UnauthorizedResponse(
                                    message="Operation cancelled"
                                )
                    elif isinstance(cmd_obj, DrainQueueCommand):
                        # Return all jobs in the queue at once
                        jobs = []
                        while self.job_queue:
                            job_id = self.job_queue.pop(0)
                            jobs.append(
                                {
                                    "job_id": job_id,
                                    "job_data": self.jobs[job_id],
                                }
                            )
                        response = OKResponse(response_data=jobs)
                    else:
                        response = UnauthorizedResponse(
                            message="Unhandled command type"
                        )
                except Exception as e:
                    response = UnauthorizedResponse(message=str(e))

                # Send back response as JSON
                writer.write((json.dumps(response.model_dump()) + "\n").encode())
                await writer.drain()
        except Exception as e:
            response = UnauthorizedResponse(message=str(e))
            writer.write((json.dumps(response.model_dump()) + "\n").encode())
            await writer.drain()
        finally:
            writer.close()

    #
    # Direct client API methods
    # These are callable from other threads/processes to communicate with the central server.
    #

    async def send_job(self, job_id: str, job_data: AppMessageTypes) -> BaseResponse:
        """
        Send a job to the broker server and wait for acknowledgement.
        """
        cmd = SendJobCommand(
            job_id=job_id,
            job_data=b64encode(pickle.dumps(job_data)).decode(),
            auth_key=self.auth_key,
        )
        response = await self._send_message(self.host, self.port, cmd)
        if isinstance(response, UnauthorizedResponse):
            raise BrokerAuthenticationError(response.message)
        return response

    async def send_response(
        self, job_id: str, response_data: TResponse
    ) -> BaseResponse:
        """
        Send a response for a job to the broker server.
        """
        cmd = SendResponseCommand(
            job_id=job_id,
            response_data=b64encode(pickle.dumps(response_data)).decode(),
            auth_key=self.auth_key,
        )
        response = await self._send_message(self.host, self.port, cmd)
        if isinstance(response, UnauthorizedResponse):
            raise BrokerAuthenticationError(response.message)
        return response

    async def get_response(self, job_id: str) -> Any:
        """
        Get a response for a job from the broker server.
        If the response is not available, wait for it.
        """
        cmd = GetResponseCommand(job_id=job_id, auth_key=self.auth_key)
        response = await self._send_message(self.host, self.port, cmd)
        if isinstance(response, UnauthorizedResponse):
            raise BrokerAuthenticationError(response.message)
        if not isinstance(response, OKResponse):
            raise ValueError(f"Failed to get response: {response}")
        if response.response_data is None:
            raise ValueError("No response data")
        return pickle.loads(b64decode(response.response_data))

    async def get_job(self) -> tuple[str, AppMessageTypes]:
        """
        Get the next available job from the broker server.
        If no jobs are available, wait for one.

        Returns:
            A dictionary containing 'job_id' and 'job_data' keys
        """
        cmd = GetJobCommand(auth_key=self.auth_key)
        response = await self._send_message(self.host, self.port, cmd)
        if isinstance(response, UnauthorizedResponse):
            raise BrokerAuthenticationError(response.message)
        if not isinstance(response, OKResponse):
            raise ValueError(f"Failed to get job: {response}")
        if response.response_data is None:
            raise ValueError("No job data")
        return response.response_data["job_id"], pickle.loads(
            b64decode(response.response_data["job_data"])
        )

    async def send_and_get_response(self, job_data: AppMessageTypes) -> Any:
        """
        Send a job to the broker server and wait for a response. Will raise on a client
        error to break out of the current loop.

        """
        job_id = str(uuid4())
        await self.send_job(job_id, job_data)
        response = await self.get_response(job_id)
        if isinstance(response, ErrorResponse):
            raise BrokerExecutionError(response.exception, response.traceback)
        return response

    async def drain_queue(self) -> list[tuple[str, AppMessageTypes]]:
        """
        Process all pending messages in the queue until it's empty.
        Returns immediately with all currently queued messages.

        Returns:
            A list of tuples containing (job_id, job_data) for all jobs that were in the queue.
        """
        cmd = DrainQueueCommand(auth_key=self.auth_key)
        response = await self._send_message(self.host, self.port, cmd)
        if isinstance(response, UnauthorizedResponse):
            raise BrokerAuthenticationError(response.message)
        if not isinstance(response, OKResponse):
            raise ValueError(f"Failed to drain queue: {response}")
        if response.response_data is None:
            return []

        return [
            (job["job_id"], pickle.loads(b64decode(job["job_data"])))
            for job in response.response_data
        ]

    async def _send_message(
        self, host: str, port: int, message_obj: BaseCommand
    ) -> BaseResponse:
        """
        Helper function to send a command message and receive the response asynchronously.
        """
        reader, writer = await asyncio.open_connection(host, port)
        try:
            json_msg = json.dumps(message_obj.model_dump()) + "\n"
            LOGGER.debug(f"Sending message to broker: {json_msg}")
            writer.write(json_msg.encode())
            await writer.drain()

            response_line = await reader.readline()
            response_dict = json.loads(response_line.decode())
            LOGGER.debug(f"Received server response: {response_dict}")

            return cast(
                BaseResponse, response_type_adapter.validate_python(response_dict)
            )
        finally:
            writer.close()
            await writer.wait_closed()

    @classmethod
    @asynccontextmanager
    async def start_server(cls):
        server = cls()
        server.start()

        # Wait for the server to start.
        while not server.is_running:
            await asyncio.sleep(0.1)

        try:
            yield (
                server,
                BrokerServerConfig(
                    host=server.host, port=server.port, auth_key=server.auth_key
                ),
            )
        finally:
            await server.stop_server()
            server.join(timeout=5.0)  # Wait for thread to finish with timeout
            if server.is_alive():
                LOGGER.warning("Warning: Server thread did not shut down cleanly")

    @classmethod
    @asynccontextmanager
    async def new_client(cls, config: BrokerServerConfig):
        client = cls(host=config.host, port=config.port, auth_key=config.auth_key)
        yield client



from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel

from mountaineer.__tests__.client_builder.interface_builders.common import (
    create_action_wrapper,
    create_controller_wrapper,
    create_model_wrapper,
)
from mountaineer.actions.fields import FunctionActionType
from mountaineer.client_builder.interface_builders.controller import ControllerInterface
from mountaineer.client_builder.parser import (
    FieldWrapper,
)
from mountaineer.client_builder.types import Or


# Test Models
class SimpleResponse(BaseModel):
    message: str
    timestamp: datetime


class ComplexResponse(BaseModel):
    data: dict[str, Any]
    status: bool
    metadata: Optional[dict[str, str]] = None


class FormData(BaseModel):
    name: str
    email: str
    preferences: dict[str, bool] = {}


class TestBasicInterfaceGeneration:
    def test_simple_controller_interface(self):
        health_check = create_action_wrapper(
            "health_check", response_model=SimpleResponse
        )

        controller = create_controller_wrapper(
            "ApiBaseController", actions={"health_check": health_check}
        )

        interface = ControllerInterface.from_controller(controller)
        ts_code = interface.to_js()

        assert "export interface" in ts_code
        assert "ApiBaseController" in ts_code
        assert "health_check" in ts_code
        assert "SimpleResponse" in ts_code

    def test_controller_with_multiple_actions(self):
        get_resource = create_action_wrapper(
            "get_resource",
            params=[FieldWrapper("id", str, True)],
            response_model=SimpleResponse,
        )

        update_resource = create_action_wrapper(
            "update_resource",
            request_body=create_model_wrapper(FormData, "FormData"),
            response_model=ComplexResponse,
            action_type=FunctionActionType.SIDEEFFECT,
        )

        controller = create_controller_wrapper(
            "ResourceController",
            actions={"get_resource": get_resource, "update_resource": update_resource},
        )

        interface = ControllerInterface.from_controller(controller)
        ts_code = interface.to_js()

        assert "get_resource" in ts_code
        assert "update_resource" in ts_code
        assert "params: {" in ts_code
        assert "FormData" in ts_code


class TestInheritanceHandling:
    def test_single_inheritance(self):
        # Create base controller
        base_controller = create_controller_wrapper(
            "ApiBaseController",
            actions={
                "health_check": create_action_wrapper(
                    "health_check", response_model=SimpleResponse
                )
            },
        )

        # Create resource controller inheriting from base
        resource_controller = create_controller_wrapper(
            "ResourceController",
            actions={
                "get_resource": create_action_wrapper(
                    "get_resource", response_model=SimpleResponse
                )
            },
            superclasses=[base_controller],
        )

        interface = ControllerInterface.from_controller(resource_controller)

        assert "ApiBaseController" in interface.include_superclasses
        assert "extends ApiBaseController" in interface.to_js()

    def test_multi_level_inheritance(self):
        # Create the inheritance chain
        base_controller = create_controller_wrapper("ApiBaseController")
        resource_controller = create_controller_wrapper(
            "ResourceController", superclasses=[base_controller]
        )
        extended_controller = create_controller_wrapper(
            "ExtendedController",
            actions={
                "specialized_action": create_action_wrapper(
                    "specialized_action", response_model=ComplexResponse
                )
            },
            superclasses=[resource_controller],
        )

        interface = ControllerInterface.from_controller(extended_controller)

        # Only the immediate superclasses should be specified
        assert "ResourceController" in interface.include_superclasses
        assert "extends ResourceController" in interface.to_js()
        assert "ApiBaseController" not in interface.to_js()


class TestParameterHandling:
    def test_optional_parameters(self):
        action = create_action_wrapper(
            "optional_action",
            params=[
                FieldWrapper("param1", Or(str, None), False),
                FieldWrapper("param2", int, False),
            ],
            response_model=SimpleResponse,
        )

        controller = create_controller_wrapper(
            "OptionalParamsController", actions={"optional_action": action}
        )

        interface = ControllerInterface.from_controller(controller)
        ts_code = interface.to_js()

        assert "params?: {" in ts_code
        assert "param1?: string" in ts_code
        assert "param2?: number" in ts_code

    def test_required_parameters(self):
        action = create_action_wrapper(
            "required_action",
            params=[FieldWrapper("required_param", str, True)],
            response_model=SimpleResponse,
        )

        controller = create_controller_wrapper(
            "RequiredParamsController", actions={"required_action": action}
        )

        interface = ControllerInterface.from_controller(controller)
        ts_code = interface.to_js()

        assert "params: {" in ts_code
        assert "required_param: string" in ts_code



import shutil
import subprocess
import sys
from json import dump as json_dump
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from mountaineer.__tests__.fixtures import get_fixture_path

# from mountaineer.development.manager import DevAppManager

AppPackageType = tuple[str, Path, Path]


def create_package_json(views_path: Path) -> None:
    """
    Create a package.json file with necessary React dependencies.

    """
    package_json = {
        "name": "test-package",
        "version": "1.0.0",
        "description": "Test package for mountaineer",
        "main": "index.js",
        "scripts": {"test": 'echo "Error: no test specified" && exit 1'},
        "dependencies": {"react": "^18.2.0", "react-dom": "^18.2.0"},
        "devDependencies": {
            "@types/react": "^18.2.0",
            "@types/react-dom": "^18.2.0",
            "typescript": "^5.0.0",
        },
    }

    with open(views_path / "package.json", "w") as f:
        json_dump(package_json, f, indent=2)


def setup_npm_environment(package_path: Path) -> None:
    """
    Install npm dependencies in the package directory.

    """
    subprocess.run(
        ["npm", "install"],
        cwd=package_path,
        check=True,
        capture_output=True,
        text=True,
    )


@pytest.fixture(scope="module")
def simple_package_dependencies():
    """
    Cache the results of create_package_json() system-wide so we don't have to
    re-install node modules for each test that needs to execute against our
    Javascript environment.

    """
    # The tmp_path fixture is scoped to the individual test function level, so we need
    # a manual cache for module scoped fixtures
    with TemporaryDirectory() as tmp_path_raw:
        tmp_path = Path(tmp_path_raw)

        # Create our standard package.json in this path
        create_package_json(tmp_path)

        # Install dependencies
        setup_npm_environment(tmp_path)

        yield tmp_path / "node_modules"


@pytest.fixture
def isolated_package_dir(
    tmp_path: Path,
    simple_package_dependencies: Path,
    request,
):
    """
    Create test package structure with unique name per test so we allow
    client functions to modify their files without adverse affects on other tests.

    Provides:
    - An isolated python package directory
    - A views directory with node_modules copied in
    - A package.json in the views directory

    """
    test_name = request.node.name.replace("test_", "")
    pkg_name = f"test_package_{test_name}".replace("[", "_").replace("]", "_")

    # Create the python code directory
    pkg_dir = tmp_path / pkg_name
    pkg_dir.mkdir()

    # Create a views directory
    views_dir = pkg_dir / "views"
    views_dir.mkdir()

    # Create a package.json in the views directory
    create_package_json(views_dir)
    shutil.copytree(simple_package_dependencies, views_dir / "node_modules")

    # Make the path reachable only within this test scope
    sys.path.insert(0, str(tmp_path))
    try:
        yield pkg_dir, pkg_name
    finally:
        sys.path.pop(0)


@pytest.fixture
def app_package(isolated_package_dir: tuple[Path, str]) -> AppPackageType:
    """
    A simple AppController, with a single component controller. Sets up a complete
    React environment with necessary dependencies.

    """
    package_path, package_name = isolated_package_dir

    # Package init
    (package_path / "__init__.py").touch()

    # Views directory with TypeScript React component
    views_dir = package_path / "views"

    # Set up package.json and install dependencies
    create_package_json(views_dir)
    setup_npm_environment(views_dir)

    # Controller
    controller_file = package_path / "test_controller.py"
    controller_file.write_text(
        (get_fixture_path("mock_webapp") / "simple_controller.py").read_text()
    )

    (views_dir / "test_controller").mkdir()
    (views_dir / "test_controller" / "page.tsx").write_text("")

    return package_name, package_path, controller_file



from pathlib import Path

from create_mountaineer_app.enums import PackageManager
from create_mountaineer_app.generation import ProjectMetadata, format_template
from create_mountaineer_app.templates import get_template_path


def test_path_url_replacement():
    metadata = ProjectMetadata(
        project_name="TEST_PROJECT_NAME",
        author_name="TEST_AUTHOR",
        author_email="TEST_EMAIL",
        package_manager=PackageManager.POETRY,
        use_tailwind=True,
        editor_config=None,
        create_stub_files=True,
        project_path=Path("fake-path"),
        mountaineer_min_version="0.1.0",
        mountaineer_dev_path=None,
    )
    project_template_base = get_template_path("project")
    bundle = format_template(
        project_template_base / "[project_name]/app.py", project_template_base, metadata
    )
    assert bundle.path == "TEST_PROJECT_NAME/app.py"



from pathlib import Path
from re import sub as re_sub

import pytest

from mountaineer import mountaineer as mountaineer_rs  # type: ignore
from mountaineer.__tests__.fixtures import get_fixture_path
from mountaineer.client_compiler.source_maps import (
    SourceMapParser,
    SourceMapSchema,
)
from mountaineer.test_utilities import benchmark_function


@pytest.mark.asyncio
# TODO: Lower threshold once we can selectively run benchmark tests with --release
@benchmark_function(0.7, time_budget_seconds=10)
async def test_parse_source_map_parse(
    start_timing,
    end_timing,
):
    parser = SourceMapParser(get_fixture_path("home_controller_source_map.js.map"))
    start_timing()
    parser.parse()
    end_timing()
    assert parser.source_map

    source_filenames = {
        Path(source_path).name for source_path in parser.source_map.sources
    }
    assert source_filenames == {
        "react.development.js",
        "index.js",
        "scheduler.development.js",
        "react-dom.development.js",
        "client.js",
        "synthetic_client.tsx",
        "live_reload.ts",
        "page.tsx",
        "useServer.ts",
        "api.ts",
        "links.ts",
        "actions.ts",
    }
    assert parser.get_original_location(37, 13)


def test_map_exception():
    exception_stacktrace = """
    Example client error
        at Page (<anonymous>:12882:13)
        at renderWithHooks (<anonymous>:6333:26)

    """

    def make_map_metadata(
        line_number: int, column_number: int, source_line: int, source_column: int
    ) -> mountaineer_rs.MapMetadata:
        """
        Workaround to deal with rust's MapMetadata constructor
        only accepting line number and column number.

        """
        metadata = mountaineer_rs.MapMetadata(
            line_number=line_number,
            column_number=column_number,
        )
        metadata.source_index = 0
        metadata.source_line = source_line
        metadata.source_column = source_column
        metadata.symbol_index = 0
        return metadata

    parser = SourceMapParser("")
    parser.source_map = SourceMapSchema(
        version=3,
        sources=["test.ts"],
        names=["symbol_test1", "symbol_test2"],
        mappings="",
        sourceRoot="",
        file="test.js",
    )
    parser.parsed_mappings = {
        (12882, 13): make_map_metadata(
            line_number=12882,
            column_number=13,
            source_line=500,
            source_column=10,
        ),
        (6333, 26): make_map_metadata(
            line_number=6333,
            column_number=13,
            source_line=600,
            source_column=20,
        ),
    }

    assert re_sub(r"\s+", "", parser.map_exception(exception_stacktrace)) == re_sub(
        r"\s+",
        "",
        """
    Example client error
        at Page (:500:10)
        at renderWithHooks (:600:20)
    """,
    )



{% if create_stub_files %}
from mountaineer import sideeffect, ControllerBase, Metadata, RenderBase
from iceaxe.mountaineer import DatabaseDependencies
from iceaxe import DBConnection, select

from fastapi import Depends

from {{project_name}} import models


class HomeRender(RenderBase):
    items: list[models.DetailItem]


class HomeController(ControllerBase):
    url = "/"
    view_path = "/app/home/page.tsx"

    async def render(
        self,
        session: DBConnection = Depends(DatabaseDependencies.get_db_connection)
    ) -> HomeRender:
        items = await session.exec(select(models.DetailItem))
        return HomeRender(
            items=items,
            metadata=Metadata(title="Home"),
        )

    @sideeffect
    async def new_detail(
        self,
        session: DBConnection = Depends(DatabaseDependencies.get_db_connection)
    ) -> None:
        obj = models.DetailItem(description="Untitled Item")
        await session.insert([obj])
{% endif %}
